#!/usr/bin/env python
#
# NEED POSIX (i.e. *Cygwin* on Windows).
"""
Script to bump, commit and tag new versions.

USAGE:
  bumpver
  bumpver [-n] [-f] [-c] [-a] [-t <message>]  <new-ver>

Without <new-ver> prints version extracted from current file.
Don't add a 'v' prefix!

OPTIONS:
  -a, --amend       Amend current commit for setting the "chore(ver): ..." msg.
  -f, --force       Bump (and optionally) commit/tag if version exists/is same.
  -n, --dry-run     Do not write files - just pretend.
  -c, --commit      Commit afterwardswith a commit-message describing version bump.
  -t, --tag=<msg>   Adds a signed tag with the given message (commit implied).


- Pre-releases: when working on some version
    X.YbN               # Beta release
    X.YrcN  or  X.YcN   # Release Candidate
    X.Y                 # Final release
- Post-release:
    X.YaN.postM         # Post-release of an alpha release
    X.YrcN.postM        # Post-release of a release candidate
- Dev-release:
    X.YaN.devM          # Developmental release of an alpha release
    X.Y.postN.devM      # Developmental release of a post-release

EXAMPLE:
    bumpver -t 'Mostly model changes' 1.6.2b0

"""

import functools as fnt
import os.path as osp
import re
import sys
from datetime import datetime

import docopt

my_dir = osp.dirname(__file__)

VFILE = osp.join(my_dir, "..", "pypiserver", "__init__.py")
VFILE_regex_version = re.compile(r'version *= *__version__ *= *"([^"]+)"')
VFILE_regex_datetime = re.compile(r'__updated__ *= *"([^"]+)"')
VFILE_regex_date = re.compile(r'__updated__ *= *"([^"\s]+)\s')

RFILE = osp.join(my_dir, "..", "README.md")

PYTEST_ARGS = [osp.join("tests", "test_docs.py")]


class CmdException(Exception):
    pass


def get_current_date_info() -> (str, str):
    now = datetime.now()
    new_datetime = now.strftime("%Y-%m-%d %H:%M:%S%z")
    new_date = now.strftime("%Y-%m-%d")
    return (new_datetime, new_date)


@fnt.lru_cache()
def read_txtfile(fpath):
    with open(fpath, "rt", encoding="utf-8") as fp:
        return fp.read()


def extract_file_regexes(fpath, regexes):
    """
    :param regexes:
        A sequence of regexes to "search", having a single capturing-group.
    :return:
        One groups per regex, or raise if any regex did not match.
    """
    txt = read_txtfile(fpath)
    matches = [regex.search(txt) for regex in regexes]

    if not all(matches):
        raise CmdException(
            "Failed extracting current versions with: %s"
            "\n  matches: %s" % (regexes, matches)
        )

    return [m.group(1) for m in matches]


def replace_substrings(files, subst_pairs):
    for fpath in files:
        txt = read_txtfile(fpath)

        replacements = []
        for old, new in subst_pairs:
            replacements.append((old, new, txt.count(old)))
            txt = txt.replace(old, new)

        yield (txt, fpath, replacements)


def format_syscmd(cmd):
    if isinstance(cmd, (list, tuple)):
        cmd = " ".join('"%s"' % s if " " in s else s for s in cmd)
    else:
        assert isinstance(cmd, str), cmd

    return cmd


def strip_ver2_commonprefix(ver1, ver2):
    cprefix = osp.commonprefix([ver1, ver2])
    if cprefix:
        striplen = cprefix.rfind(".")
        if striplen > 0:
            striplen += 1
        else:
            striplen = len(cprefix)
        ver2 = ver2[striplen:]

    return ver2


def run_testcases():
    import pytest

    retcode = pytest.main(PYTEST_ARGS)
    if retcode:
        raise CmdException(
            "Doc TCs failed(%s), probably version-bumping has failed!" % retcode
        )


def exec_cmd(cmd):
    import subprocess as sbp

    err = sbp.call(cmd, stderr=sbp.STDOUT)
    if err:
        raise CmdException("Failed(%i) on: %s" % (err, format_syscmd(cmd)))


def do_commit(new_ver, old_ver, dry_run, amend, ver_files):
    import pathlib

    cmt_msg = "chore(ver): bump %s-->%s" % (old_ver, new_ver)

    ver_files = [pathlib.Path(f).as_posix() for f in ver_files]
    git_add = ["git", "add"] + ver_files
    git_cmt = ["git", "commit", "-m", cmt_msg]
    if amend:
        git_cmt.append("--amend")
    commands = [git_add, git_cmt]

    for cmd in commands:
        cmd_str = format_syscmd(cmd)
        if dry_run:
            yield "DRYRUN: %s" % cmd_str
        else:
            yield "EXEC: %s" % cmd_str
            exec_cmd(cmd)


def do_tag(tag, tag_msg, dry_run, force):
    cmd = ["git", "tag", tag, "-s", "-m", tag_msg]
    if force:
        cmd.append("--force")
    cmd_str = format_syscmd(cmd)
    if dry_run:
        yield "DRYRUN: %s" % cmd_str
    else:
        yield "EXEC: %s" % cmd_str
        exec_cmd(cmd)


def bumpver(
    new_ver, dry_run=False, force=False, amend=False, tag_name_or_commit=None
):
    """
    :param tag_name_or_commit:
        if true, do `git commit`, if string, also `git tag` with that as msg.
    """
    if amend:
        ## Restore previous version before extracting it.
        cmd = "git checkout HEAD~ --".split()
        cmd.append(VFILE)
        cmd.append(RFILE)
        exec_cmd(cmd)

    regexes = [VFILE_regex_version, VFILE_regex_datetime, VFILE_regex_date]
    old_ver, old_datetime, old_date = extract_file_regexes(VFILE, regexes)

    if not new_ver:
        yield old_ver
        yield old_datetime
        yield old_date
    else:
        if new_ver == old_ver:
            msg = "Version '%s'already bumped"
            if force:
                msg += ", but --force  effected."
                yield msg % new_ver
            else:
                msg += "!\n Use of --force recommended."
                raise CmdException(msg % new_ver)

        new_datetime, new_date = get_current_date_info()
        ver_files = [osp.normpath(f) for f in [VFILE, RFILE]]
        subst_pairs = [
            (old_ver, new_ver),
            (old_datetime, new_datetime),
            (old_date, new_date),
        ]

        for repl in replace_substrings(ver_files, subst_pairs):
            new_txt, fpath, replacements = repl

            if not dry_run:
                with open(fpath, "wt", encoding="utf-8") as fp:
                    fp.write(new_txt)

            yield "%s: " % fpath
            for old, new, nrepl in replacements:
                yield "  %i x (%24s --> %s)" % (nrepl, old, new)

        yield "...now launching DocTCs..."
        run_testcases()

        if tag_name_or_commit is not None:
            yield from do_commit(new_ver, old_ver, dry_run, amend, ver_files)

            if isinstance(tag_name_or_commit, str):
                tag = "v%s" % new_ver
                yield from do_tag(tag, tag_name_or_commit, dry_run, force)


def main(*args):
    opts = docopt.docopt(__doc__, argv=args)

    new_ver = opts["<new-ver>"]

    assert not new_ver or new_ver[0] != "v", (
        "Version '%s' must NOT start with `v`!" % new_ver
    )

    commit = opts["--commit"]
    tag = opts["--tag"]
    if tag:
        tag_name_or_commit = tag
    elif commit:
        tag_name_or_commit = True
    else:
        tag_name_or_commit = None

    try:
        for i in bumpver(
            new_ver,
            opts["--dry-run"],
            opts["--force"],
            opts["--amend"],
            tag_name_or_commit,
        ):
            print(i)
    except CmdException as ex:
        sys.exit(str(ex))
    except Exception as ex:
        print("Unexpected error happened.")
        raise ex


if __name__ == "__main__":
    main(*sys.argv[1:])



import abc
import functools
import hashlib
import itertools
import logging
import os
import typing as t
from pathlib import Path

from .cache import CacheManager, ENABLE_CACHING
from .core import PkgFile
from .pkg_helpers import (
    normalize_pkgname,
    is_listed_path,
    guess_pkgname_and_version,
)

if t.TYPE_CHECKING:
    from .config import _ConfigCommon as Configuration


log = logging.getLogger(__name__)


PathLike = t.Union[str, os.PathLike]


class IBackend(abc.ABC):
    @abc.abstractmethod
    def get_all_packages(self) -> t.Iterable[PkgFile]:
        pass

    @abc.abstractmethod
    def find_project_packages(self, project: str) -> t.Iterable[PkgFile]:
        pass

    @abc.abstractmethod
    def find_version(self, name: str, version: str) -> t.Iterable[PkgFile]:
        pass

    @abc.abstractmethod
    def get_projects(self) -> t.Iterable[str]:
        pass

    @abc.abstractmethod
    def exists(self, filename: str) -> bool:
        pass

    @abc.abstractmethod
    def digest(self, pkg: PkgFile) -> t.Optional[str]:
        pass

    @abc.abstractmethod
    def package_count(self) -> int:
        pass

    @abc.abstractmethod
    def add_package(self, filename: str, stream: t.BinaryIO) -> None:
        pass

    @abc.abstractmethod
    def remove_package(self, pkg: PkgFile) -> None:
        pass


class Backend(IBackend, abc.ABC):
    def __init__(self, config: "Configuration"):
        self.hash_algo = config.hash_algo

    @abc.abstractmethod
    def get_all_packages(self) -> t.Iterable[PkgFile]:
        """Implement this method to return an Iterable of all packages (as
        PkgFile objects) that are available in the Backend.
        """
        pass

    @abc.abstractmethod
    def add_package(self, filename: str, stream: t.BinaryIO) -> None:
        """Add a package to the Backend. `filename` is the package's filename
        (without any directory parts). It is just a name, there is no file by
        that name (yet). `stream` is an open file-like object that can be used
        to read the file's content. To convert the package into an actual file
        on disk, run `write_file(filename, stream)`.
        """
        pass

    @abc.abstractmethod
    def remove_package(self, pkg: PkgFile) -> None:
        """Remove a package from the Backend"""
        pass

    @abc.abstractmethod
    def exists(self, filename: str) -> bool:
        """Does a package by the given name exist?"""
        pass

    def digest(self, pkg: PkgFile) -> t.Optional[str]:
        if self.hash_algo is None or pkg.fn is None:
            return None
        return digest_file(pkg.fn, self.hash_algo)

    def package_count(self) -> int:
        """Return a count of all available packages. When implementing a Backend
        class, either use this method as is, or override it with a more
        performant version.
        """
        return sum(1 for _ in self.get_all_packages())

    def get_projects(self) -> t.Iterable[str]:
        """Return an iterable of all (unique) projects available in the store
        in their PEP503 normalized form. When implementing a Backend class,
        either use this method as is, or override it with a more performant
        version.
        """
        return set(package.pkgname_norm for package in self.get_all_packages())

    def find_project_packages(self, project: str) -> t.Iterable[PkgFile]:
        """Find all packages from a given project. The project may be given
        as either the normalized or canonical name. When implementing a
        Backend class, either use this method as is, or override it with a
        more performant version.
        """
        return (
            x
            for x in self.get_all_packages()
            if normalize_pkgname(project) == x.pkgname_norm
        )

    def find_version(self, name: str, version: str) -> t.Iterable[PkgFile]:
        """Return all packages that match PkgFile.pkgname == name and
        PkgFile.version == version` When implementing a Backend class,
        either use this method as is, or override it with a more performant
        version.
        """
        return filter(
            lambda pkg: pkg.pkgname == name and pkg.version == version,
            self.get_all_packages(),
        )


class SimpleFileBackend(Backend):
    def __init__(self, config: "Configuration"):
        super().__init__(config)
        self.roots = [Path(root).resolve() for root in config.roots]

    def get_all_packages(self) -> t.Iterable[PkgFile]:
        return itertools.chain.from_iterable(listdir(r) for r in self.roots)

    def add_package(self, filename: str, stream: t.BinaryIO) -> None:
        write_file(stream, self.roots[0].joinpath(filename))

    def remove_package(self, pkg: PkgFile) -> None:
        if pkg.fn is not None:
            try:
                os.remove(pkg.fn)
            except FileNotFoundError:
                log.warning(
                    "Tried to remove %s, but it is already gone", pkg.fn
                )
            except OSError:
                log.exception("Unexpected error removing package: %s", pkg.fn)
                raise

    def exists(self, filename: str) -> bool:
        return any(
            filename == existing_file.name
            for root in self.roots
            for existing_file in all_listed_files(root)
        )


class CachingFileBackend(SimpleFileBackend):
    def __init__(
        self,
        config: "Configuration",
        cache_manager: t.Optional[CacheManager] = None,
    ):
        super().__init__(config)

        self.cache_manager = cache_manager or CacheManager()  # type: ignore

    def add_package(self, filename: str, stream: t.BinaryIO) -> None:
        super().add_package(filename, stream)
        self.cache_manager.invalidate_root_cache(self.roots[0])

    def remove_package(self, pkg: PkgFile) -> None:
        super().remove_package(pkg)
        self.cache_manager.invalidate_root_cache(pkg.root)

    def get_all_packages(self) -> t.Iterable[PkgFile]:
        return itertools.chain.from_iterable(
            self.cache_manager.listdir(r, listdir) for r in self.roots
        )

    def digest(self, pkg: PkgFile) -> t.Optional[str]:
        if self.hash_algo is None or pkg.fn is None:
            return None
        return self.cache_manager.digest_file(
            pkg.fn, self.hash_algo, digest_file
        )


def write_file(fh: t.BinaryIO, destination: PathLike) -> None:
    """write a byte stream into a destination file. Writes are chunked to reduce
    the memory footprint
    """
    chunk_size = 2**20  # 1 MB
    offset = fh.tell()
    try:
        with open(destination, "wb") as dest:
            for chunk in iter(lambda: fh.read(chunk_size), b""):
                dest.write(chunk)
    finally:
        fh.seek(offset)


def listdir(root: Path) -> t.Iterator[PkgFile]:
    root = root.resolve()
    files = all_listed_files(root)
    yield from valid_packages(root, files)


def all_listed_files(root: Path) -> t.Iterator[Path]:
    for dirpath, dirnames, filenames in os.walk(root):
        dirnames[:] = (
            dirname for dirname in dirnames if is_listed_path(Path(dirname))
        )
        for filename in filenames:
            if not is_listed_path(Path(filename)):
                continue
            filepath = root / dirpath / filename
            if Path(filepath).is_file():
                yield filepath


def valid_packages(root: Path, files: t.Iterable[Path]) -> t.Iterator[PkgFile]:
    for file in files:
        res = guess_pkgname_and_version(str(file.name))
        if res is not None:
            pkgname, version = res
            fn = str(file)
            root_name = str(root)
            yield PkgFile(
                pkgname=pkgname,
                version=version,
                fn=fn,
                root=root_name,
                relfn=fn[len(root_name) + 1 :],
            )


def digest_file(file_path: PathLike, hash_algo: str) -> str:
    """
    Reads and digests a file according to specified hashing-algorith.

    :param file_path: path to a file on disk
    :param hash_algo: any algo contained in :mod:`hashlib`
    :return: <hash_algo>=<hex_digest>

    From http://stackoverflow.com/a/21565932/548792
    """
    blocksize = 2**16
    digester = hashlib.new(hash_algo)
    with open(file_path, "rb") as f:
        for block in iter(lambda: f.read(blocksize), b""):
            digester.update(block)
    return f"{hash_algo}={digester.hexdigest()}"


def get_file_backend(config: "Configuration") -> Backend:
    if ENABLE_CACHING:
        return CachingFileBackend(config)
    return SimpleFileBackend(config)


PkgFunc = t.TypeVar("PkgFunc", bound=t.Callable[..., t.Iterable[PkgFile]])


def with_digester(func: PkgFunc) -> PkgFunc:
    @functools.wraps(func)
    def add_digester_method(
        self: "BackendProxy", *args: t.Any, **kwargs: t.Any
    ) -> t.Iterable[PkgFile]:
        packages = func(self, *args, **kwargs)
        for package in packages:
            package.digester = self.backend.digest
            yield package

    return t.cast(PkgFunc, add_digester_method)


class BackendProxy(IBackend):
    def __init__(self, wraps: Backend):
        self.backend = wraps

    @with_digester
    def get_all_packages(self) -> t.Iterable[PkgFile]:
        return self.backend.get_all_packages()

    @with_digester
    def find_project_packages(self, project: str) -> t.Iterable[PkgFile]:
        return self.backend.find_project_packages(project)

    def find_version(self, name: str, version: str) -> t.Iterable[PkgFile]:
        return self.backend.find_version(name, version)

    def get_projects(self) -> t.Iterable[str]:
        return self.backend.get_projects()

    def exists(self, filename: str) -> bool:
        assert "/" not in filename
        return self.backend.exists(filename)

    def package_count(self) -> int:
        return self.backend.package_count()

    def add_package(self, filename: str, fh: t.BinaryIO) -> None:
        assert "/" not in filename
        return self.backend.add_package(filename, fh)

    def remove_package(self, pkg: PkgFile) -> None:
        return self.backend.remove_package(pkg)

    def digest(self, pkg: PkgFile) -> t.Optional[str]:
        return self.backend.digest(pkg)



"""Management operations for pypiserver."""

from __future__ import absolute_import, print_function, unicode_literals

import itertools
import json
import os
import sys
from distutils.version import LooseVersion
from pathlib import Path
from subprocess import call
from urllib.error import URLError
from urllib.request import Request, urlopen

import pip
from packaging.version import parse as packaging_parse

from .backend import listdir
from .core import PkgFile
from .pkg_helpers import normalize_pkgname, parse_version


def is_stable_version(pversion):
    for x in ("*c", "*@", "*b"):
        if x in pversion:
            return False
    for x in pversion:
        if x.startswith("*final"):
            return True
        if x.startswith("*"):
            return False
    return False


def filter_stable_releases(releases):
    for pkg in releases:
        if is_stable_version(pkg.parsed_version):
            yield pkg


def filter_latest_pkgs(pkgs):
    pkgname2latest = {}

    for x in pkgs:
        pkgname = normalize_pkgname(x.pkgname)

        if pkgname not in pkgname2latest:
            pkgname2latest[pkgname] = x
        elif x.parsed_version > pkgname2latest[pkgname].parsed_version:
            pkgname2latest[pkgname] = x

    return pkgname2latest.values()


def build_releases(pkg, versions):
    for x in versions:
        parsed_version = parse_version(x)
        if parsed_version > pkg.parsed_version:
            yield PkgFile(pkgname=pkg.pkgname, version=x, replaces=pkg)


def get_package_releases(pkgname):
    pypi_json_content_type = "application/vnd.pypi.simple.v1+json"

    req = Request(
        f"https://pypi.org/simple/{pkgname}",
        headers={"Accept": pypi_json_content_type},
    )
    try:
        with urlopen(req) as resp:
            parsed = json.load(resp)
            return parsed["versions"]
    except URLError as error:
        # TODO(tech-debt): use a proper logging approach
        print(type(error), error.reason, sep=": ")
        return None


def find_updates(pkgset, stable_only=True):
    no_releases = set()
    filter_releases = filter_stable_releases if stable_only else (lambda x: x)

    def write(s):
        sys.stdout.write(s)
        sys.stdout.flush()

    latest_pkgs = frozenset(filter_latest_pkgs(pkgset))

    sys.stdout.write(
        f"checking {len(latest_pkgs)} packages for newer version\n"
    )
    need_update = set()

    for count, pkg in enumerate(latest_pkgs):
        if count % 40 == 0:
            write("\n")

        pypi_versions = get_package_releases(pkg.pkgname)
        if pypi_versions:
            releases = filter_releases(build_releases(pkg, pypi_versions))
            status = "."
            try:
                need_update.add(max(releases, key=lambda x: x.parsed_version))
                status = "u"
            except ValueError:
                pass
        else:
            status = "e"
            no_releases.add(pkg.pkgname)

        write(status)

    write("\n\n")

    if no_releases:
        sys.stdout.write(
            f"no releases found on pypi for"
            f" {', '.join(sorted(no_releases))}\n\n"
        )

    return need_update


class PipCmd:
    """Methods for generating pip commands."""

    @staticmethod
    def update_root(pip_version):
        """Yield an appropriate root command depending on pip version.

        Use `pip install` for `pip` 9 or lower, and `pip download` otherwise.
        """
        legacy_pip = packaging_parse(pip_version).major < 10
        pip_command = "install" if legacy_pip else "download"
        for part in ("pip", "-q", pip_command):
            yield part

    @staticmethod
    def update(
        cmd_root,
        destdir,
        pkg_name,
        pkg_version,
        index="https://pypi.org/simple",
    ):
        """Yield an update command for pip."""
        for part in cmd_root:
            yield part
        for part in ("--no-deps", "-i", index, "-d", destdir):
            yield part
        yield "{}=={}".format(pkg_name, pkg_version)


def update_package(pkg, destdir, dry_run=False):
    """Print and optionally execute a package update."""
    print(
        f"# update {pkg.pkgname} from {pkg.replaces.version} to {pkg.version}"
    )

    cmd = tuple(
        PipCmd.update(
            PipCmd.update_root(pip.__version__),
            destdir or os.path.dirname(pkg.replaces.fn),
            pkg.pkgname,
            pkg.version,
        )
    )

    print(" ".join(cmd), end="\n\n")
    if not dry_run:
        call(cmd)


def update(pkgset, destdir=None, dry_run=False, stable_only=True):
    """Print and optionally execute pip update commands.

    :param pkgset: the set of currently available packages
    :param str destdir: the destination directory for downloads
    :param dry_run: whether commands should be executed (rather than
        just being printed)
    :param stable_only: whether only stable (non prerelease) updates
        should be considered.
    """
    need_update = find_updates(pkgset, stable_only=stable_only)
    for pkg in sorted(need_update, key=lambda x: x.pkgname):
        update_package(pkg, destdir, dry_run=dry_run)


def update_all_packages(
    roots, destdir=None, dry_run=False, stable_only=True, ignorelist=None
):
    all_packages = itertools.chain.from_iterable(
        listdir(Path(r)) for r in roots
    )

    skip_packages = set(ignorelist or ())

    packages = frozenset(
        [pkg for pkg in all_packages if pkg.pkgname not in skip_packages]
    )

    update(packages, destdir, dry_run, stable_only)



#
# The cache implementation is only used when the watchdog package
# is installed
#

from os.path import dirname
from pathlib import Path
import typing as t
import threading

try:
    from watchdog.observers import Observer

    ENABLE_CACHING = True

except ImportError:
    Observer = None

    ENABLE_CACHING = False

if t.TYPE_CHECKING:
    from pypiserver.core import PkgFile


class CacheManager:
    """
    A naive cache implementation for listdir and digest_file

    The listdir_cache is just a giant list of PkgFile objects, and
    for simplicity it is invalidated anytime a modification occurs
    within the directory it represents. If we were smarter about
    the way that the listdir data structure were created/stored,
    then we could do more granular invalidation. In practice, this
    is good enough for now.

    The digest_cache exists on a per-file basis, because computing
    hashes on large files can get expensive, and it's very easy to
    invalidate specific filenames.
    """

    def __init__(self):
        if not ENABLE_CACHING:
            raise RuntimeError(
                "Please install the extra cache requirements by running 'pip "
                "install pypiserver[cache]' to use the CachingFileBackend"
            )

        # Cache for listdir output
        self.listdir_cache = {}

        # Cache for hashes: two-level dictionary
        # -> key: hash_algo, value: dict
        #    -> key: file path, value: hash
        # We assume that the hash_algo value will never be erased
        self.digest_cache = {}

        self.observer = Observer()
        self.observer.start()

        # Directories being watched
        self.watched = set()

        self.watch_lock = threading.Lock()
        self.digest_lock = threading.Lock()
        self.listdir_lock = threading.Lock()

    def listdir(
        self,
        root: t.Union[Path, str],
        impl_fn: t.Callable[[Path], t.Iterable["PkgFile"]],
    ) -> t.Iterable["PkgFile"]:
        root = str(root)
        with self.listdir_lock:
            try:
                return self.listdir_cache[root]
            except KeyError:
                # check to see if we're watching
                with self.watch_lock:
                    if root not in self.watched:
                        self._watch(root)

                v = list(impl_fn(Path(root)))
                self.listdir_cache[root] = v
                return v

    def digest_file(
        self, fpath: str, hash_algo: str, impl_fn: t.Callable[[str, str], str]
    ) -> str:
        with self.digest_lock:
            try:
                cache = self.digest_cache[hash_algo]
            except KeyError:
                cache = self.digest_cache.setdefault(hash_algo, {})

            try:
                return cache[fpath]
            except KeyError:
                root = dirname(fpath)
                with self.watch_lock:
                    if root not in self.watched:
                        self._watch(root)

            # TODO: move this outside of the lock... but there's not a good
            #       way to do this without a race condition if the file
            #       gets modified
            v = impl_fn(fpath, hash_algo)
            cache[fpath] = v
            return v

    def _watch(self, root: str):
        self.watched.add(root)
        self.observer.schedule(_EventHandler(self, root), root, recursive=True)

    def invalidate_root_cache(self, root: t.Union[Path, str]):
        with self.listdir_lock:
            self.listdir_cache.pop(str(root), None)


class _EventHandler:
    def __init__(self, cache: CacheManager, root: str):
        self.cache = cache
        self.root = root

    def dispatch(self, event):
        """Called by watchdog observer"""
        cache = self.cache

        # Don't care about directory events
        if event.is_directory:
            return

        # Lazy: just invalidate the whole cache
        cache.invalidate_root_cache(self.root)

        # Digests are more expensive: invalidate specific paths
        paths = []

        if event.event_type == "moved":
            paths.append(event.src_path)
            paths.append(event.dest_path)
        else:
            paths.append(event.src_path)

        with cache.digest_lock:
            for _, subcache in cache.digest_cache.items():
                for path in paths:
                    subcache.pop(path, None)



""" NOT YET IMPLEMENTED

Plugins are callable setuptools entrypoints that are invoked at startup that
a developer may use to extend the behaviour of pypiserver. A plugin for example
may add an additional Backend to the system. A plugin will be called
with the following keyword arguments

* app: the Bottle App object
* add_argument: A callable for registering command line arguments for your
    plugin using the argparse cli library
* backends: A Dict[str, callable] object that you may register a backend to.
    The key is the identifier for the backend in the `--backend` command line
    argument.
    The callable must take a single argument `config` as a Configuration object
    and return a Backend instance. It may be the class constructor or a factory
    function to construct a Backend object

In the future, the plugin callable may be called with additional keyword
arguments, so a plugin should accept a **kwargs variadic keyword argument.
"""

from pypiserver.backend import SimpleFileBackend, CachingFileBackend
from pypiserver import get_file_backend

DEFAULT_PACKAGE_DIRECTORIES = ["~/packages"]


# register this as a setuptools entrypoint under the 'pypiserver.plugin' key
def my_plugin(add_argument, backends, **_):
    add_argument(
        "package_directory",
        default=DEFAULT_PACKAGE_DIRECTORIES,
        nargs="*",
        help="The directory from which to serve packages.",
    )
    backends.update(
        {
            "auto": get_file_backend,
            "simple-dir": SimpleFileBackend,
            "cached-dir": CachingFileBackend,
        }
    )



import logging
import mimetypes
import os
import re
import xml.dom.minidom
import xmlrpc.client as xmlrpclib
import zipfile
from collections import defaultdict
from collections import namedtuple
from io import BytesIO
from json import dumps
from urllib.parse import urljoin, urlparse

from pypiserver.config import RunConfig
from . import __version__
from . import core
from .bottle import (
    static_file,
    redirect,
    request,
    response,
    HTTPError,
    Bottle,
    template,
)
from .pkg_helpers import guess_pkgname_and_version, normalize_pkgname_for_url

log = logging.getLogger(__name__)
config: RunConfig
app = Bottle()


class auth:
    """decorator to apply authentication if specified for the decorated method & action"""

    def __init__(self, action):
        self.action = action

    def __call__(self, method):
        def protector(*args, **kwargs):
            if self.action in config.authenticate:
                if not request.auth or request.auth[1] is None:
                    raise HTTPError(
                        401, headers={"WWW-Authenticate": 'Basic realm="pypi"'}
                    )
                if not config.auther(*request.auth):
                    raise HTTPError(403)
            return method(*args, **kwargs)

        return protector


@app.hook("before_request")
def log_request():
    log.info(config.log_req_frmt, request.environ)


@app.hook("before_request")
def print_request():
    parsed = urlparse(request.urlparts.scheme + "://" + request.urlparts.netloc)
    request.custom_host = parsed.netloc
    request.custom_fullpath = (
        parsed.path.rstrip("/") + "/" + request.fullpath.lstrip("/")
    )


@app.hook("after_request")
def log_response():
    log.info(
        config.log_res_frmt,
        {  # vars(response))  ## DOES NOT WORK!
            "response": response,
            "status": response.status,
            "headers": response.headers,
            "body": response.body,
            "cookies": response._cookies,
        },
    )


@app.error
def log_error(http_error):
    log.info(config.log_err_frmt, vars(http_error))


@app.route("/favicon.ico")
def favicon():
    return HTTPError(404)


@app.route("/")
def root():
    fp = request.custom_fullpath

    # Ensure template() does not consider `msg` as filename!
    msg = config.welcome_msg + "\n"
    return template(
        msg,
        URL=request.url.rstrip("/") + "/",
        VERSION=__version__,
        NUMPKGS=config.backend.package_count(),
        PACKAGES=fp.rstrip("/") + "/packages/",
        SIMPLE=fp.rstrip("/") + "/simple/",
    )


_bottle_upload_filename_re = re.compile(r"^[a-z0-9_.!+-]+$", re.I)


def is_valid_pkg_filename(fname):
    """See https://github.com/pypiserver/pypiserver/issues/102"""
    return _bottle_upload_filename_re.match(fname) is not None


def doc_upload():
    try:
        content = request.files["content"]
    except KeyError:
        raise HTTPError(400, "Missing 'content' file-field!")
    zip_data = content.file.read()
    try:
        zf = zipfile.ZipFile(BytesIO(zip_data))
        zf.getinfo("index.html")
    except Exception:
        raise HTTPError(400, "not a zip file")


def remove_pkg():
    name = request.forms.get("name")
    version = request.forms.get("version")
    if not name or not version:
        msg = f"Missing 'name'/'version' fields: name={name}, version={version}"
        raise HTTPError(400, msg)

    pkgs = list(config.backend.find_version(name, version))
    if not pkgs:
        raise HTTPError(404, f"{name} ({version}) not found")
    for pkg in pkgs:
        config.backend.remove_package(pkg)


Upload = namedtuple("Upload", "pkg sig")


def file_upload():
    ufiles = Upload._make(
        request.files.get(f, None) for f in ("content", "gpg_signature")
    )
    if not ufiles.pkg:
        raise HTTPError(400, "Missing 'content' file-field!")
    if (
        ufiles.sig
        and f"{ufiles.pkg.raw_filename}.asc" != ufiles.sig.raw_filename
    ):
        raise HTTPError(
            400,
            f"Unrelated signature {ufiles.sig!r} for package {ufiles.pkg!r}!",
        )

    for uf in ufiles:
        if not uf:
            continue
        if (
            not is_valid_pkg_filename(uf.raw_filename)
            or guess_pkgname_and_version(uf.raw_filename) is None
        ):
            raise HTTPError(400, f"Bad filename: {uf.raw_filename}")

        if not config.overwrite and config.backend.exists(uf.raw_filename):
            log.warning(
                f"Cannot upload {uf.raw_filename!r} since it already exists! \n"
                "  You may start server with `--overwrite` option. "
            )
            raise HTTPError(
                409,
                f"Package {uf.raw_filename!r} already exists!\n"
                "  You may start server with `--overwrite` option.",
            )

        config.backend.add_package(uf.raw_filename, uf.file)
        if request.auth:
            user = request.auth[0]
        else:
            user = "anon"
        log.info(f"User {user!r} stored {uf.raw_filename!r}.")


@app.post("/")
@auth("update")
def update():
    try:
        action = request.forms[":action"]
    except KeyError:
        raise HTTPError(400, "Missing ':action' field!")

    if action in ("verify", "submit"):
        log.warning(f"Ignored ':action': {action}")
    elif action == "doc_upload":
        doc_upload()
    elif action == "remove_pkg":
        remove_pkg()
    elif action == "file_upload":
        file_upload()
    else:
        raise HTTPError(400, f"Unsupported ':action' field: {action}")

    return ""


@app.route("/simple")
@app.route("/simple/:project")
@app.route("/packages")
@auth("list")
def pep_503_redirects(project=None):
    return redirect(request.custom_fullpath + "/", 301)


@app.post("/RPC2")
@auth("list")
def handle_rpc():
    """Handle pip-style RPC2 search requests"""
    parser = xml.dom.minidom.parse(request.body)
    methodname = (
        parser.getElementsByTagName("methodName")[0]
        .childNodes[0]
        .wholeText.strip()
    )
    log.debug(f"Processing RPC2 request for '{methodname}'")
    if methodname == "search":
        value = (
            parser.getElementsByTagName("string")[0]
            .childNodes[0]
            .wholeText.strip()
        )
        response = []
        ordering = 0
        for p in config.backend.get_all_packages():
            if p.pkgname.count(value) > 0:
                # We do not presently have any description/summary, returning
                # version instead
                d = {
                    "_pypi_ordering": ordering,
                    "version": p.version,
                    "name": p.pkgname,
                    "summary": p.version,
                }
                response.append(d)
            ordering += 1
        call_string = xmlrpclib.dumps(
            (response,), "search", methodresponse=True
        )
        return call_string


@app.route("/simple/")
@auth("list")
def simpleindex():
    links = sorted(config.backend.get_projects())
    tmpl = """<!DOCTYPE html>
<html lang="en">
    <head>
        <meta charset="utf-8">
        <meta name="viewport" content="width=device-width, initial-scale=1">
        <title>Simple Index</title>
    </head>
    <body>
        <h1>Simple Index</h1>
        % for p in links:
                <a href="{{p}}/">{{p}}</a><br>
        % end
    </body>
</html>
    """
    return template(tmpl, links=links)


@app.route("/simple/:project/")
@auth("list")
def simple(project):
    # PEP 503: require normalized project
    normalized = normalize_pkgname_for_url(project)
    if project != normalized:
        return redirect(f"/simple/{normalized}/", 301)

    packages = sorted(
        config.backend.find_project_packages(project),
        key=lambda x: (x.parsed_version, x.relfn),
    )
    if not packages:
        if not config.disable_fallback:
            return redirect(f"{config.fallback_url.rstrip('/')}/{project}/")
        return HTTPError(404, f"Not Found ({normalized} does not exist)\n\n")

    current_uri = request.custom_fullpath

    links = (
        (
            os.path.basename(pkg.relfn),
            urljoin(current_uri, f"../../packages/{pkg.fname_and_hash}"),
        )
        for pkg in packages
    )

    tmpl = """<!DOCTYPE html>
<html lang="en">
    <head>
        <meta charset="utf-8">
        <meta name="viewport" content="width=device-width, initial-scale=1">
        <title>Links for {{project}}</title>
    </head>
    <body>
        <h1>Links for {{project}}</h1>
        % for file, href in links:
            <a href="{{href}}">{{file}}</a><br>
        % end
    </body>
</html>
    """
    return template(tmpl, project=project, links=links)


@app.route("/packages/")
@auth("list")
def list_packages():
    fp = request.custom_fullpath
    packages = sorted(
        config.backend.get_all_packages(),
        key=lambda x: (os.path.dirname(x.relfn), x.pkgname, x.parsed_version),
    )

    links = (
        (pkg.relfn_unix, urljoin(fp, pkg.fname_and_hash)) for pkg in packages
    )

    tmpl = """<!DOCTYPE html>
<html lang="en">
    <head>
        <meta charset="utf-8">
        <meta name="viewport" content="width=device-width, initial-scale=1">
        <title>Index of packages</title>
    </head>
    <body>
        <h1>Index of packages</h1>
        % for file, href in links:
            <a href="{{href}}">{{file}}</a><br>
        % end
    </body>
</html>
    """
    return template(tmpl, links=links)


@app.route("/packages/:filename#.*#")
@auth("download")
def server_static(filename):
    entries = config.backend.get_all_packages()
    for x in entries:
        f = x.relfn_unix
        if f == filename:
            response = static_file(
                filename,
                root=x.root,
                mimetype=mimetypes.guess_type(filename)[0],
            )
            if config.cache_control:
                response.set_header(
                    "Cache-Control", f"public, max-age={config.cache_control}"
                )
            return response

    return HTTPError(404, f"Not Found ({filename} does not exist)\n\n")


@app.route("/:project/json")
@auth("list")
def json_info(project):
    # PEP 503: require normalized project
    normalized = normalize_pkgname_for_url(project)
    if project != normalized:
        return redirect(f"/{normalized}/json", 301)

    packages = sorted(
        config.backend.find_project_packages(project),
        key=lambda x: x.parsed_version,
        reverse=True,
    )

    if not packages:
        raise HTTPError(404, f"package {project} not found")

    latest_version = packages[0].version
    releases = defaultdict(list)
    req_url = request.url
    for x in packages:
        releases[x.version].append(
            {"url": urljoin(req_url, "../../packages/" + x.relfn)}
        )

    rv = {"info": {"version": latest_version}, "releases": releases}
    response.content_type = "application/json"
    return dumps(rv)


@app.route("/:project")
@app.route("/:project/")
def bad_url(project):
    """Redirect unknown root URLs to /simple/."""
    return redirect(core.get_bad_url_redirect_path(request, project))



"""Default gunicorn config for the docker environment.

To override, mount a new gunicorn config at /data/gunicorn.conf.py in your
Docker container.
"""

# pylint: disable=invalid-name

# Enable to log every request
# accesslog = "-"
errorlog = "-"
preload_app = True
workers = 1
worker_class = "gevent"



#! /usr/bin/env py.test
"""
Checks an actual pypi-server against various clients.

The tests below are using 3 ways to startup pypi-servers:

- "open": a per-module server instance without any authed operations,
  serving a single `wheel` package, on a fixed port.
- "open": a per-module server instance with authed 'download/upload'
  operations, serving a single `wheel` package, on a fixed port.
- "new_server": starting a new server with any configurations on each test.

"""
import contextlib
import itertools
import os
import shutil
import socket
import re
import sys
import time
import typing as t
from collections import namedtuple
from pathlib import Path
from shlex import split
from subprocess import Popen
from urllib.error import URLError
from urllib.request import urlopen

import pytest

# ######################################################################
# Fixtures & Helper Functions
# ######################################################################


CURRENT_PATH = Path(__file__).parent
ports = itertools.count(10000)
Srv = namedtuple("Srv", ("port", "root"))


@contextlib.contextmanager
def run_server(root, authed=False, other_cli=""):
    """Run a server, optionally with partial auth enabled."""
    htpasswd = (
        CURRENT_PATH.joinpath("../fixtures/htpasswd.a.a").expanduser().resolve()
    )
    pswd_opt_choices = {
        True: f"-P {htpasswd} -a update,download",
        False: "-P. -a.",
        "partial": f"-P {htpasswd} -a update",
    }
    pswd_opts = pswd_opt_choices[authed]

    port = next(ports)
    cmd = (
        f"{sys.executable} -m pypiserver.__main__ "
        f"run -vvv --overwrite -i 127.0.0.1 "
        f"-p {port} {pswd_opts} {other_cli} {root}"
    )
    proc = Popen(cmd.split(), bufsize=2**16)
    srv = Srv(port, root)
    try:
        wait_until_ready(srv)
        assert proc.poll() is None
        yield srv
    finally:
        print(f"Killing {srv}")
        _kill_proc(proc)


def wait_until_ready(srv: Srv, n_tries=10):
    for _ in range(n_tries):
        if is_ready(srv):
            return True
        time.sleep(0.5)
    raise TimeoutError


def is_ready(srv: Srv):
    try:
        return urlopen(build_url(srv.port), timeout=0.5).getcode() in (
            200,
            401,
        )
    except (URLError, socket.timeout):
        return False


def _kill_proc(proc):
    proc.terminate()
    try:
        proc.wait(timeout=1)
    finally:
        proc.kill()


def build_url(port: t.Union[int, str], user: str = "", pswd: str = "") -> str:
    auth = f"{user}:{pswd}@" if user or pswd else ""
    return f"http://{auth}localhost:{port}"


def run_setup_py(path: Path, arguments: str) -> int:
    return os.system(f"{sys.executable} {path / 'setup.py'} {arguments}")


def run_py_build(srcdir: Path, flags: str) -> int:
    return os.system(f"{sys.executable} -m build {flags} {srcdir}")


# A test-distribution to check if
#    bottle supports uploading 100's of packages,
#    see: https://github.com/pypiserver/pypiserver/issues/82
#
# Has been run once `pip wheel .`, just to generate:
#    ./wheelhouse/centodeps-0.0.0-cp34-none-win_amd64.whl
#
SETUP_PY = """\
from setuptools import setup

setup(
    name="centodeps",
    install_requires=["a==1.0"] * 200,
    options={
        "bdist_wheel": {"universal": True},
    },
    packages=[],
)
"""


@pytest.fixture(scope="module")
def project(tmp_path_factory):
    projdir = tmp_path_factory.mktemp("project") / "centodeps"
    projdir.mkdir(parents=True, exist_ok=True)
    projdir.joinpath("setup.py").write_text(SETUP_PY)
    return projdir


@pytest.fixture(scope="session")
def server_root(tmp_path_factory):
    return tmp_path_factory.mktemp("root")


@pytest.fixture(scope="module")
def wheel_file(project, tmp_path_factory):
    distdir = tmp_path_factory.mktemp("dist")
    if re.match("^3\.7", sys.version):
        assert run_setup_py(project, f"bdist_wheel -d {distdir}") == 0
    else:
        assert (
            run_py_build(project, f"--wheel --no-isolation --outdir {distdir}")
            == 0
        )
    wheels = list(distdir.glob("centodeps*.whl"))
    assert len(wheels) > 0
    return wheels[0]


@pytest.fixture()
def hosted_wheel_file(wheel_file, server_root):
    dst = server_root / wheel_file.name
    shutil.copy(wheel_file, dst)
    yield dst
    if dst.is_file():
        dst.unlink()


def clear_directory(root: Path):
    for path in root.iterdir():
        if path.is_file():
            path.unlink()


@pytest.fixture(scope="module")
def _open_server(server_root):
    with run_server(server_root, authed=False) as srv:
        yield srv


@pytest.fixture
def open_server(_open_server: Srv):
    yield _open_server
    clear_directory(_open_server.root)


@pytest.fixture(scope="module")
def _authed_server(server_root):
    with run_server(server_root, authed=True) as srv:
        yield srv


@pytest.fixture
def authed_server(_authed_server):
    yield _authed_server
    clear_directory(_authed_server.root)


@pytest.fixture(scope="module")
def _partial_auth_server(server_root):
    with run_server(server_root, authed="partial") as srv:
        yield srv


@pytest.fixture
def partial_authed_server(_partial_auth_server):
    yield _partial_auth_server
    clear_directory(_partial_auth_server.root)


@pytest.fixture
def empty_packdir(tmp_path_factory):
    return tmp_path_factory.mktemp("dists")


def pip_download(
    cmd: str,
    port: t.Union[int, str],
    install_dir: str,
    user: str = None,
    pswd: str = None,
) -> int:
    url = build_url(port, user, pswd)
    return _run_pip(f"-vv download -d {install_dir} -i {url} {cmd}")


def _run_pip(cmd: str) -> int:
    ncmd = (
        "pip --no-cache-dir --disable-pip-version-check "
        f"--retries 0 --timeout 5 --no-input {cmd}"
    )
    print(f"PIP: {ncmd}")
    proc = Popen(split(ncmd))
    proc.communicate()
    return proc.returncode


@pytest.fixture
def pipdir(tmp_path_factory):
    return tmp_path_factory.mktemp("pip")


@contextlib.contextmanager
def pypirc_file(repo, username="''", password="''"):
    pypirc_path = Path.home() / ".pypirc"
    old_pypirc = pypirc_path.read_text() if pypirc_path.is_file() else None
    pypirc_path.write_text(
        "\n".join(
            (
                "[distutils]",
                "index-servers: test",
                "",
                "[test]",
                f"repository: {repo}",
                f"username: {username}",
                f"password: {password}",
            )
        )
    )
    try:
        yield pypirc_path
    finally:
        if old_pypirc:
            pypirc_path.write_text(old_pypirc)
        else:
            pypirc_path.unlink()


@pytest.fixture
def open_pypirc(open_server):
    with pypirc_file(repo=build_url(open_server.port)) as path:
        yield path


@pytest.fixture
def authed_pypirc(authed_server):
    username, password = "a", "a"
    with pypirc_file(
        repo=build_url(authed_server.port),
        username=username,
        password=password,
    ) as path:
        yield path


def run_twine(command: str, package: str, conf: str) -> None:
    proc = Popen(
        split(
            f"twine {command} --repository test --config-file {conf} {package}"
        )
    )
    proc.communicate()
    assert not proc.returncode, f"Twine {command} failed. See stdout/err"


# ######################################################################
# Tests
# ######################################################################

all_servers = [
    ("open_server", "open_pypirc"),
    ("authed_server", "authed_pypirc"),
    ("partial_authed_server", "authed_pypirc"),
]


def test_pip_install_package_not_found(open_server, pipdir):
    assert pip_download("centodeps", open_server.port, pipdir) != 0
    assert not list(pipdir.iterdir())


def test_pip_install_open_succeeds(open_server, hosted_wheel_file, pipdir):
    assert pip_download("centodeps", open_server.port, pipdir) == 0
    assert pipdir.joinpath(hosted_wheel_file.name).is_file()


@pytest.mark.usefixtures("wheel_file")
def test_pip_install_authed_fails(authed_server, pipdir):
    assert pip_download("centodeps", authed_server.port, pipdir) != 0
    assert not list(pipdir.iterdir())


def test_pip_install_authed_succeeds(authed_server, hosted_wheel_file, pipdir):
    assert (
        pip_download(
            "centodeps", authed_server.port, pipdir, user="a", pswd="a"
        )
        == 0
    )
    assert pipdir.joinpath(hosted_wheel_file.name).is_file()


def test_partial_authed_open_download(partial_authed_server):
    """Validate that partial auth still allows downloads."""
    url = build_url(partial_authed_server.port) + "/simple"
    resp = urlopen(url)
    assert resp.getcode() == 200


@pytest.mark.parametrize("hash_algo", ("md5", "sha256", "sha512"))
@pytest.mark.usefixtures("hosted_wheel_file")
def test_hash_algos(server_root, pipdir, hash_algo):
    """Test twine upload with no authentication"""
    with run_server(
        server_root, other_cli="--hash-algo {}".format(hash_algo)
    ) as srv:
        assert pip_download("centodeps", srv.port, pipdir) == 0


@pytest.mark.parametrize(["server_fixture", "pypirc_fixture"], all_servers)
def test_twine_upload(
    server_fixture, pypirc_fixture, server_root, wheel_file, request
):
    """Test twine upload with no authentication"""
    assert len(list(server_root.iterdir())) == 0
    request.getfixturevalue(server_fixture)
    pypirc = request.getfixturevalue(pypirc_fixture)

    run_twine("upload", wheel_file, conf=pypirc)

    assert len(list(server_root.iterdir())) == 1
    assert server_root.joinpath(wheel_file.name).is_file(), (
        wheel_file.name,
        list(server_root.iterdir()),
    )


@pytest.mark.parametrize(["server_fixture", "pypirc_fixture"], all_servers)
def test_twine_register(server_fixture, pypirc_fixture, wheel_file, request):
    """Test unauthenticated twine registration"""
    request.getfixturevalue(server_fixture)
    pypirc = request.getfixturevalue(pypirc_fixture)
    run_twine("register", wheel_file, conf=pypirc)



"""
Test module for app initialization
"""

# Standard library imports
import logging
import pathlib
import typing as t

# Third party imports
import pytest


# Local imports
import pypiserver
from pypiserver.config import Config

logger = logging.getLogger(__name__)

TEST_DIR = pathlib.Path(__file__).parent
HTPASS_FILE = TEST_DIR / "../fixtures/htpasswd.a.a"
WELCOME_FILE = TEST_DIR / "sample_msg.html"


# TODO: make these tests meaningful
@pytest.mark.parametrize(
    "conf_options",
    [
        {},
        {"root": "~/stable_packages"},
        {
            "root": "~/unstable_packages",
            "authenticated": "upload",
            "passwords": str(HTPASS_FILE),
        },
        # Verify that the strip parser works properly.
        {"authenticated": str("upload")},
    ],
)
def test_paste_app_factory(conf_options: dict) -> None:
    """Test the paste_app_factory method"""
    pypiserver.paste_app_factory({}, **conf_options)  # type: ignore


def test_app_factory() -> None:
    assert pypiserver.app() is not pypiserver.app()


@pytest.mark.parametrize(
    "incoming, updated",
    (
        (
            {"authenticated": []},
            {"authenticate": []},
        ),
        (
            {"passwords": "./foo"},
            {"password_file": "./foo"},
        ),
        (
            {"root": str(TEST_DIR)},
            {"roots": [TEST_DIR.expanduser().resolve()]},
        ),
        (
            {"root": [str(TEST_DIR), str(TEST_DIR)]},
            {
                "roots": [
                    TEST_DIR.expanduser().resolve(),
                    TEST_DIR.expanduser().resolve(),
                ]
            },
        ),
        (
            {"redirect_to_fallback": False},
            {"disable_fallback": True},
        ),
        (
            {"server": "auto"},
            {"server_method": "auto"},
        ),
        (
            {"welcome_file": str(WELCOME_FILE.resolve())},
            {"welcome_msg": WELCOME_FILE.read_text()},
        ),
    ),
)
def test_backwards_compat_kwargs_conversion(
    incoming: t.Dict[str, t.Any], updated: t.Dict[str, t.Any]
) -> None:
    """Test converting legacy kwargs to modern ones."""
    assert pypiserver.backwards_compat_kwargs(incoming) == updated


@pytest.mark.parametrize(
    "kwargs",
    (
        {"redirect_to_fallback": False, "disable_fallback": False},
        {"disable_fallback": False, "redirect_to_fallback": False},
    ),
)
def test_backwards_compat_kwargs_duplicate_check(
    kwargs: t.Dict[str, t.Any]
) -> None:
    """Duplicate legacy and modern kwargs cause an error."""
    with pytest.raises(ValueError) as err:
        pypiserver.backwards_compat_kwargs(kwargs)
    assert "('redirect_to_fallback', 'disable_fallback')" in str(err.value)


def test_setup_routes_from_config_customized_endpoint() -> None:
    _app = pypiserver.setup_routes_from_config(
        pypiserver.app(),
        Config.default_with_overrides(health_endpoint="/healthz"),
    )
    assert "/healthz" in (route.rule for route in _app.routes)


def test_setup_routes_from_config_invalid_customized_endpoint() -> None:
    with pytest.raises(RuntimeError, match="overlaps with existing routes"):
        pypiserver.setup_routes_from_config(
            pypiserver.app(),
            Config.default_with_overrides(health_endpoint="/simple"),
        )



"""Tests for config parsing."""

import hashlib
import typing as t
import itertools
import pathlib
import sys

import pytest

from pypiserver.backend import SimpleFileBackend, BackendProxy
from pypiserver.config import DEFAULTS, Config, RunConfig, UpdateConfig

FILE_DIR = pathlib.Path(__file__).parent.resolve()

# Username and password stored in the htpasswd.a.a test file.
HTPASS_TEST_FILE = str(FILE_DIR / "../fixtures/htpasswd.a.a")
HTPASS_TEST_USER = "a"
HTPASS_TEST_PASS = "a"

TEST_WELCOME_FILE = str(pathlib.Path(__file__).parent / "sample_msg.html")
TEST_IGNORELIST_FILE = str(pathlib.Path(__file__).parent / "test-ignorelist")


class ConfigTestCase(t.NamedTuple):
    # A description of the test case
    case: str
    # Arguments to pass to the Config constructor
    args: t.List[str]
    # Legacy arguments that should yield an equivalent config class
    legacy_args: t.List[str]
    # The config class the arguments should resolve to
    exp_config_type: t.Type
    # Expected values in the config. These don't necessarily need to be
    # exclusive. Instead, they should just look at the attributes relevant
    # to the test case at hand. A special "_test" key, if present, should
    # map to a function that takes the config as an argument. If this
    # returns a falsey value, the test will be failed.
    exp_config_values: t.Dict[str, t.Any]


# The iterables generated by this function are designed to be unpacked
# into the _CONFIG_TEST_PARAMS constant.
def generate_subcommand_test_cases(
    case: str,
    extra_args: t.List[str] = None,
    exp_config_values: t.Dict[str, t.Any] = None,
) -> t.Iterable[ConfigTestCase]:
    """Generate `run` and `update` test cases automatically.

    Use to avoid having to specify individual cases for situations like
    global arguments, where the resultant configs should have the same values.

    These tests also help to ensure parity between legacy and modern orderings,
    since generally the only difference between the two should be the presence
    or absence of the subcommand.

    :param case: the test case name. will be combined with the subcommand
        to generate a case name for the resultant case.
    :param extra_args: arguments to pass after the subcommand positional
        arguments
    :param extra_legacy_args: legacy arguments to pass in addition to the
        subcommand arguments, if any
    :param exp_config_values: the values that should be present on both
        run and update test cases.
    """
    extra_args = extra_args or []
    extra_legacy_args = extra_args
    exp_config_values = exp_config_values or {}
    # The legacy "update" subcommand was specified with an optional `-U`
    # argument. This allows us to map the subcommand to that argument, so
    # we can include it in the resulting legacy args.
    legacy_base_arg_map = {"update": ["-U"]}
    # A mapping of subcommands to their expected Config types.
    config_type_map = {
        "run": RunConfig,
        "update": UpdateConfig,
    }
    return (
        ConfigTestCase(
            case="{subcmd}: {case}",
            args=[subcmd, *extra_args],
            legacy_args=[
                *legacy_base_arg_map.get(subcmd, []),
                *extra_legacy_args,
            ],
            exp_config_type=config_type_map[subcmd],
            exp_config_values=exp_config_values,
        )
        for subcmd in ("run", "update")
    )


# Define Config test parameters
_CONFIG_TEST_PARAMS: t.Tuple[ConfigTestCase, ...] = (
    # ******************************************************************
    # Raw subcommands
    # ******************************************************************
    *generate_subcommand_test_cases(
        case="no arguments",
    ),
    # ******************************************************************
    # Global args
    # ******************************************************************
    # Package directories
    *generate_subcommand_test_cases(
        case="no package directory specified",
        exp_config_values={"roots": DEFAULTS.PACKAGE_DIRECTORIES},
    ),
    *generate_subcommand_test_cases(
        case="single package directory specified",
        extra_args=[str(FILE_DIR)],
        exp_config_values={"roots": [FILE_DIR]},
    ),
    *generate_subcommand_test_cases(
        case="multiple package directory specified",
        extra_args=[str(FILE_DIR), str(FILE_DIR.parent)],
        exp_config_values={
            "roots": [
                FILE_DIR,
                FILE_DIR.parent,
            ]
        },
    ),
    ConfigTestCase(
        case="update with package directory (out-of-order legacy order)",
        args=["update", str(FILE_DIR)],
        legacy_args=[str(FILE_DIR), "-U"],
        exp_config_type=UpdateConfig,
        exp_config_values={"roots": [FILE_DIR]},
    ),
    ConfigTestCase(
        case="update with multiple package directories (weird ordering)",
        args=["update", str(FILE_DIR), str(FILE_DIR.parent)],
        legacy_args=[str(FILE_DIR), "-U", str(FILE_DIR.parent)],
        exp_config_type=UpdateConfig,
        exp_config_values={
            "roots": [
                FILE_DIR,
                FILE_DIR.parent,
            ]
        },
    ),
    # verbosity
    *(
        # Generate verbosity test-cases for 0 through 5 -v arguments,
        # for all subcommands.
        itertools.chain.from_iterable(
            # This inner iterable (generate(...) for verbosity in range(5))
            # will be an iterable of 5 items, where each item is essentially
            # an n-tuple, where n is the number of subcommands. Passing this
            # iterable to chain.from_iterable() flattens it, so it is just
            # one long iterable of cases. These are then unpacked into the
            # test case tuple with the *, above.
            generate_subcommand_test_cases(
                case=f"verbosity {verbosity}",
                extra_args=["-v" for _ in range(verbosity)],
                exp_config_values={"verbosity": verbosity},
            )
            for verbosity in range(5)
        )
    ),
    # log-file
    *generate_subcommand_test_cases(
        case="log file unspecified", exp_config_values={"log_file": None}
    ),
    *generate_subcommand_test_cases(
        case="log file specified",
        extra_args=["--log-file", "foo"],
        exp_config_values={"log_file": "foo"},
    ),
    # log-stream
    *generate_subcommand_test_cases(
        case="log stream unspecified",
        exp_config_values={"log_stream": DEFAULTS.LOG_STREAM},
    ),
    *generate_subcommand_test_cases(
        case="log stream set to stdout",
        extra_args=["--log-stream", "stdout"],
        exp_config_values={"log_stream": sys.stdout},
    ),
    *generate_subcommand_test_cases(
        case="log stream set to stderr",
        extra_args=["--log-stream", "stderr"],
        exp_config_values={"log_stream": sys.stderr},
    ),
    *generate_subcommand_test_cases(
        case="log stream set to none",
        extra_args=["--log-stream", "none"],
        exp_config_values={"log_stream": None},
    ),
    *generate_subcommand_test_cases(
        case="log format unset",
        exp_config_values={"log_frmt": DEFAULTS.LOG_FRMT},
    ),
    *generate_subcommand_test_cases(
        case="log format set",
        extra_args=["--log-frmt", "foobar %(message)s"],
        exp_config_values={"log_frmt": "foobar %(message)s"},
    ),
    # ******************************************************************
    # Run subcommand args
    # ******************************************************************
    # port
    ConfigTestCase(
        case="Run: port unspecified",
        args=["run"],
        legacy_args=[],
        exp_config_type=RunConfig,
        exp_config_values={"port": DEFAULTS.PORT},
    ),
    ConfigTestCase(
        case="Run: port specified",
        args=["run", "-p", "9900"],
        legacy_args=["-p", "9900"],
        exp_config_type=RunConfig,
        exp_config_values={"port": 9900},
    ),
    ConfigTestCase(
        case="Run: port specified (long form)",
        args=["run", "--port", "9900"],
        legacy_args=["--port", "9900"],
        exp_config_type=RunConfig,
        exp_config_values={"port": 9900},
    ),
    # interface
    ConfigTestCase(
        case="Run: interface unspecified",
        args=["run"],
        legacy_args=[],
        exp_config_type=RunConfig,
        exp_config_values={"host": DEFAULTS.INTERFACE},
    ),
    ConfigTestCase(
        case="Run: interface specified",
        args=["run", "-i", "1.1.1.1"],
        legacy_args=["-i", "1.1.1.1"],
        exp_config_type=RunConfig,
        exp_config_values={"host": "1.1.1.1"},
    ),
    ConfigTestCase(
        case="Run: interface specified (long form)",
        args=["run", "--interface", "1.1.1.1"],
        legacy_args=["--interface", "1.1.1.1"],
        exp_config_type=RunConfig,
        exp_config_values={"host": "1.1.1.1"},
    ),
    ConfigTestCase(
        case="Run: host specified",
        args=["run", "-H", "1.1.1.1"],
        legacy_args=["-H", "1.1.1.1"],
        exp_config_type=RunConfig,
        exp_config_values={"host": "1.1.1.1"},
    ),
    ConfigTestCase(
        case="Run: host specified (long form)",
        args=["run", "--host", "1.1.1.1"],
        legacy_args=["--host", "1.1.1.1"],
        exp_config_type=RunConfig,
        exp_config_values={"host": "1.1.1.1"},
    ),
    # authenticate
    ConfigTestCase(
        case="Run: authenticate unspecified",
        args=["run"],
        legacy_args=[],
        exp_config_type=RunConfig,
        exp_config_values={"authenticate": DEFAULTS.AUTHENTICATE},
    ),
    ConfigTestCase(
        case="Run: authenticate specified as non-default value",
        args=["run", "-a", "list"],
        legacy_args=["-a", "list"],
        exp_config_type=RunConfig,
        exp_config_values={"authenticate": ["list"]},
    ),
    ConfigTestCase(
        case="Run: authenticate specified with multiple values",
        args=["run", "-a", "list, update,download"],
        legacy_args=["-a", "list, update,download"],
        exp_config_type=RunConfig,
        exp_config_values={"authenticate": ["download", "list", "update"]},
    ),
    ConfigTestCase(
        case="Run: authenticate specified with dot",
        # both auth and pass must be specified as empty if one of them is empty.
        args=["run", "-a", ".", "-P", "."],
        legacy_args=["-a", ".", "-P", "."],
        exp_config_type=RunConfig,
        exp_config_values={
            "authenticate": [],
            "_test": lambda conf: bool(conf.auther("foo", "bar")) is True,
        },
    ),
    # passwords
    ConfigTestCase(
        case="Run: passwords file unspecified",
        args=["run"],
        legacy_args=[],
        exp_config_type=RunConfig,
        exp_config_values={"password_file": None},
    ),
    ConfigTestCase(
        "Run: passwords file specified",
        args=["run", "-P", HTPASS_TEST_FILE],
        legacy_args=["-P", HTPASS_TEST_FILE],
        exp_config_type=RunConfig,
        exp_config_values={
            "password_file": HTPASS_TEST_FILE,
            "_test": lambda conf: (
                bool(conf.auther("foo", "bar")) is False
                and bool(conf.auther("a", "a")) is True
            ),
        },
    ),
    ConfigTestCase(
        "Run: passwords file specified (long-form)",
        args=["run", "--passwords", HTPASS_TEST_FILE],
        legacy_args=["--passwords", HTPASS_TEST_FILE],
        exp_config_type=RunConfig,
        exp_config_values={
            "password_file": HTPASS_TEST_FILE,
            "_test": (
                lambda conf: (
                    bool(conf.auther("foo", "bar")) is False
                    and conf.auther("a", "a") is True
                )
            ),
        },
    ),
    ConfigTestCase(
        "Run: passwords file empty ('.')",
        # both auth and pass must be specified as empty if one of them is empty.
        args=["run", "-P", ".", "-a", "."],
        legacy_args=["-P", ".", "-a", "."],
        exp_config_type=RunConfig,
        exp_config_values={
            "password_file": ".",
            "_test": lambda conf: bool(conf.auther("foo", "bar")) is True,
        },
    ),
    # disable-fallback
    ConfigTestCase(
        case="Run: disable-fallback unspecified",
        args=["run"],
        legacy_args=[],
        exp_config_type=RunConfig,
        exp_config_values={"disable_fallback": False},
    ),
    ConfigTestCase(
        case="Run: disable-fallback set",
        args=["run", "--disable-fallback"],
        legacy_args=["--disable-fallback"],
        exp_config_type=RunConfig,
        exp_config_values={"disable_fallback": True},
    ),
    # fallback-url
    ConfigTestCase(
        case="Run: fallback-url unspecified",
        args=["run"],
        legacy_args=[],
        exp_config_type=RunConfig,
        exp_config_values={"fallback_url": DEFAULTS.FALLBACK_URL},
    ),
    ConfigTestCase(
        case="Run: fallback-url specified",
        args=["run", "--fallback-url", "foobar.com"],
        legacy_args=["--fallback-url", "foobar.com"],
        exp_config_type=RunConfig,
        exp_config_values={"fallback_url": "foobar.com"},
    ),
    # health-endpoint
    ConfigTestCase(
        case="Run: health-endpoint unspecified",
        args=["run"],
        legacy_args=[],
        exp_config_type=RunConfig,
        exp_config_values={"health_endpoint": DEFAULTS.HEALTH_ENDPOINT},
    ),
    ConfigTestCase(
        case="Run: health-endpoint specified",
        args=["run", "--health-endpoint", "/healthz"],
        legacy_args=["--health-endpoint", "/healthz"],
        exp_config_type=RunConfig,
        exp_config_values={"health_endpoint": "/healthz"},
    ),
    # server method
    ConfigTestCase(
        case="Run: server method unspecified",
        args=["run"],
        legacy_args=[],
        exp_config_type=RunConfig,
        exp_config_values={"server_method": DEFAULTS.SERVER_METHOD},
    ),
    *(
        ConfigTestCase(
            case="Run: server method set to {arg}",
            args=["run", "--server", arg],
            legacy_args=["--server", arg],
            exp_config_type=RunConfig,
            exp_config_values={"server_method": arg},
        )
        for arg in (
            "auto",
            "cherrypy",
            "gevent",
            "gunicorn",
            "paste",
            "twisted",
            "wsgiref",
        )
    ),
    ConfigTestCase(
        case="Run: server method is case insensitive",
        args=["run", "--server", "CherryPy"],
        legacy_args=["--server", "CherryPy"],
        exp_config_type=RunConfig,
        exp_config_values={"server_method": "cherrypy"},
    ),
    # overwrite
    ConfigTestCase(
        "Run: overwrite unset",
        args=["run"],
        legacy_args=[],
        exp_config_type=RunConfig,
        exp_config_values={"overwrite": False},
    ),
    ConfigTestCase(
        case="Run: overwrite set (long-form)",
        args=["run", "-o"],
        legacy_args=["-o"],
        exp_config_type=RunConfig,
        exp_config_values={"overwrite": True},
    ),
    ConfigTestCase(
        case="Run: overwrite set (long-form)",
        args=["run", "--overwrite"],
        legacy_args=["--overwrite"],
        exp_config_type=RunConfig,
        exp_config_values={"overwrite": True},
    ),
    # hash-algo
    ConfigTestCase(
        case="Run: hash-algo unspecified",
        args=["run"],
        legacy_args=[],
        exp_config_type=RunConfig,
        exp_config_values={"hash_algo": DEFAULTS.HASH_ALGO},
    ),
    *(
        ConfigTestCase(
            case="Run: hash-algo {}",
            args=["run", "--hash-algo", algo],
            legacy_args=["--hash-algo", algo],
            exp_config_type=RunConfig,
            exp_config_values={"hash_algo": algo},
        )
        for algo in hashlib.algorithms_available
    ),
    *(
        ConfigTestCase(
            case="Run: hash-algo disabled",
            args=["run", "--hash-algo", off_value],
            legacy_args=["--hash-algo", off_value],
            exp_config_type=RunConfig,
            exp_config_values={"hash_algo": None},
        )
        for off_value in ("0", "off", "false", "no", "NO")
    ),
    # welcome file
    ConfigTestCase(
        case="Run: welcome file unspecified",
        args=["run"],
        legacy_args=[],
        exp_config_type=RunConfig,
        exp_config_values={
            "_test": lambda conf: "Welcome to pypiserver" in conf.welcome_msg
        },
    ),
    ConfigTestCase(
        case="Run: custom welcome file specified",
        args=["run", "--welcome", TEST_WELCOME_FILE],
        legacy_args=["--welcome", TEST_WELCOME_FILE],
        exp_config_type=RunConfig,
        exp_config_values={
            "_test": lambda conf: "Hello pypiserver tester!" in conf.welcome_msg
        },
    ),
    # cache-control
    ConfigTestCase(
        case="Run: cache-control unspecified",
        args=["run"],
        legacy_args=[],
        exp_config_type=RunConfig,
        exp_config_values={"cache_control": None},
    ),
    ConfigTestCase(
        case="Run: cache-control specified",
        args=["run", "--cache-control", "1900"],
        legacy_args=["--cache-control", "1900"],
        exp_config_type=RunConfig,
        exp_config_values={"cache_control": 1900},
    ),
    # log-req-frmt
    ConfigTestCase(
        case="Run: log request format unspecified",
        args=["run"],
        legacy_args=[],
        exp_config_type=RunConfig,
        exp_config_values={"log_req_frmt": DEFAULTS.LOG_REQ_FRMT},
    ),
    ConfigTestCase(
        case="Run: log request format specified",
        args=["run", "--log-req-frmt", "foo"],
        legacy_args=["--log-req-frmt", "foo"],
        exp_config_type=RunConfig,
        exp_config_values={"log_req_frmt": "foo"},
    ),
    # log-res-frmt
    ConfigTestCase(
        case="Run: log response format unspecified",
        args=["run"],
        legacy_args=[],
        exp_config_type=RunConfig,
        exp_config_values={"log_res_frmt": DEFAULTS.LOG_RES_FRMT},
    ),
    ConfigTestCase(
        case="Run: log response format specified",
        args=["run", "--log-res-frmt", "foo"],
        legacy_args=["--log-res-frmt", "foo"],
        exp_config_type=RunConfig,
        exp_config_values={"log_res_frmt": "foo"},
    ),
    # log-err-frmt
    ConfigTestCase(
        case="Run: log error format unspecified",
        args=["run"],
        legacy_args=[],
        exp_config_type=RunConfig,
        exp_config_values={"log_err_frmt": DEFAULTS.LOG_ERR_FRMT},
    ),
    ConfigTestCase(
        case="Run: log error format specified",
        args=["run", "--log-err-frmt", "foo"],
        legacy_args=["--log-err-frmt", "foo"],
        exp_config_type=RunConfig,
        exp_config_values={"log_err_frmt": "foo"},
    ),
    # backend
    ConfigTestCase(
        "Run: backend unspecified",
        args=["run"],
        legacy_args=[],
        exp_config_type=RunConfig,
        exp_config_values={
            "backend_arg": "auto",
            "_test": (
                lambda conf: (
                    isinstance(conf.backend, BackendProxy)
                    and isinstance(conf.backend.backend, SimpleFileBackend)
                )
            ),
        },
    ),
    ConfigTestCase(
        "Run: simple backend specified",
        args=["run", "--backend", "simple-dir"],
        legacy_args=["--backend", "simple-dir"],
        exp_config_type=RunConfig,
        exp_config_values={
            "_test": (
                lambda conf: (
                    isinstance(conf.backend.backend, SimpleFileBackend)
                )
            ),
        },
    ),
    # ******************************************************************
    # Update subcommand args
    # ******************************************************************
    # execute
    ConfigTestCase(
        case="Update: execute not specified",
        args=["update"],
        legacy_args=["-U"],
        exp_config_type=UpdateConfig,
        exp_config_values={"execute": False},
    ),
    ConfigTestCase(
        case="Update: execute specified",
        args=["update", "-x"],
        legacy_args=["-U", "-x"],
        exp_config_type=UpdateConfig,
        exp_config_values={"execute": True},
    ),
    ConfigTestCase(
        case="Update: execute specified (long-form)",
        args=["update", "--execute"],
        legacy_args=["-U", "--execute"],
        exp_config_type=UpdateConfig,
        exp_config_values={"execute": True},
    ),
    # download-directory
    ConfigTestCase(
        case="Update: download-directory not specified",
        args=["update"],
        legacy_args=["-U"],
        exp_config_type=UpdateConfig,
        exp_config_values={"download_directory": None},
    ),
    ConfigTestCase(
        case="Update: download-directory specified",
        args=["update", "-d", "foo"],
        legacy_args=["-U", "-d", "foo"],
        exp_config_type=UpdateConfig,
        exp_config_values={"download_directory": "foo"},
    ),
    ConfigTestCase(
        case="Update: download-directory specified (long-form)",
        args=["update", "--download-directory", "foo"],
        legacy_args=["-U", "--download-directory", "foo"],
        exp_config_type=UpdateConfig,
        exp_config_values={"download_directory": "foo"},
    ),
    # allow-unstable
    ConfigTestCase(
        case="Update: allow-unstable not specified",
        args=["update"],
        legacy_args=["-U"],
        exp_config_type=UpdateConfig,
        exp_config_values={"allow_unstable": False},
    ),
    ConfigTestCase(
        case="Update: allow-unstable specified",
        args=["update", "-u"],
        legacy_args=["-U", "-u"],
        exp_config_type=UpdateConfig,
        exp_config_values={"allow_unstable": True},
    ),
    ConfigTestCase(
        case="Update: allow-unstable specified (long-form)",
        args=["update", "--allow-unstable"],
        legacy_args=["-U", "--allow-unstable"],
        exp_config_type=UpdateConfig,
        exp_config_values={"allow_unstable": True},
    ),
    # ignorelist-file
    ConfigTestCase(
        case="Update: ignorelist-file not specified",
        args=["update"],
        legacy_args=["-U"],
        exp_config_type=UpdateConfig,
        exp_config_values={"ignorelist": []},
    ),
    ConfigTestCase(
        case="Update: ignorelist-file specified",
        args=["update", "--ignorelist-file", TEST_IGNORELIST_FILE],
        legacy_args=["-U", "--ignorelist-file", TEST_IGNORELIST_FILE],
        exp_config_type=UpdateConfig,
        exp_config_values={"ignorelist": ["mypiserver", "something"]},
    ),
    ConfigTestCase(
        case="Update: blacklist-file specified",
        args=["update", "--blacklist-file", TEST_IGNORELIST_FILE],
        legacy_args=["-U", "--blacklist-file", TEST_IGNORELIST_FILE],
        exp_config_type=UpdateConfig,
        exp_config_values={"ignorelist": ["mypiserver", "something"]},
    ),
)

# Split case names out from cases to use as pytest IDs.
# pylint: disable=unsubscriptable-object
CONFIG_TEST_PARAMS = (i[1:] for i in _CONFIG_TEST_PARAMS)
# pylint: enable=unsubscriptable-object
CONFIG_TEST_IDS = (i.case for i in _CONFIG_TEST_PARAMS)


class ConfigErrorCase(t.NamedTuple):
    """Configuration arguments that should cause errors.

    The cases include a case descrpition, a list of arguments,
    and, if desired, expected text that should be part of what
    is printed out to stderr. If no text is provided, the content
    of stderr will not be checked.
    """

    case: str
    args: t.List[str]
    exp_txt: t.Optional[str]


_CONFIG_ERROR_CASES = (
    *(
        ConfigErrorCase(
            case=f"Invalid hash algo: {val}",
            args=["run", "--hash-algo", val],
            exp_txt=f"Hash algorithm '{val}' is not available",
        )
        for val in ("true", "foo", "1", "md6")
    ),
    *(
        ConfigErrorCase(
            case=f"Invalid health endpoint: {val}",
            args=["run", "--health-endpoint", val],
            exp_txt="Invalid path for the health endpoint",
        )
        for val in ("/", "health", "/health!", "/:health", "/health?check=True")
    ),
)
# pylint: disable=unsubscriptable-object
CONFIG_ERROR_PARAMS = (i[1:] for i in _CONFIG_ERROR_CASES)
# pylint: enable=unsubscriptable-object
CONFIG_ERROR_IDS = (i.case for i in _CONFIG_ERROR_CASES)


@pytest.mark.parametrize(
    "args, legacy_args, exp_config_type, exp_config_values",
    CONFIG_TEST_PARAMS,
    ids=CONFIG_TEST_IDS,
)
def test_config(
    args: t.List[str],
    legacy_args: t.List[str],
    exp_config_type: t.Type,
    exp_config_values: t.Dict[str, t.Any],
) -> None:
    """Validate config test cases."""
    conf = Config.from_args(args)
    conf_legacy = Config.from_args(legacy_args)

    assert isinstance(conf, exp_config_type)
    assert all(
        getattr(conf, k) == v
        for k, v in exp_config_values.items()
        if k != "_test"
    ), {
        k: (getattr(conf, k), v)
        for k, v in exp_config_values.items()
        if k != "_test" and getattr(conf, k) != v
    }

    if "_test" in exp_config_values:
        assert exp_config_values["_test"](conf)

    assert conf == conf_legacy


@pytest.mark.parametrize(
    "args, exp_txt",
    CONFIG_ERROR_PARAMS,
    ids=CONFIG_ERROR_IDS,
)
def test_config_error(
    args: t.List[str],
    exp_txt: t.Optional[str],
    capsys,
) -> None:
    """Validate error cases."""
    with pytest.raises(SystemExit):
        Config.from_args(args)
    # Unfortunately the error text is printed before the SystemExit is
    # raised, rather than being raised _with_ the systemexit, so we
    # need to capture stderr and check it for our expected text, if
    # any was specified in the test case.
    if exp_txt is not None:
        assert exp_txt in capsys.readouterr().err


def test_argv_conf():
    """Config uses argv if no args are provided."""
    orig_args = list(sys.argv)

    sys.argv = [sys.argv[0], "run", "-v", "--disable-fallback"]

    try:
        conf = Config.from_args()
        assert isinstance(conf, RunConfig)
        assert conf.verbosity == 1
        assert conf.disable_fallback is True
    finally:
        sys.argv = orig_args



import os
from pathlib import WindowsPath, PureWindowsPath

import pytest

from pypiserver.pkg_helpers import guess_pkgname_and_version, is_listed_path

files = [
    ("pytz-2012b.tar.bz2", "pytz", "2012b"),
    ("pytz-2012b.tgz", "pytz", "2012b"),
    ("pytz-2012b.ZIP", "pytz", "2012b"),
    ("pytz-2012a.zip", "pytz", "2012a"),
    ("pytz-2012b.tar.xz", "pytz", "2012b"),
    ("gevent-1.0b1.win32-py2.6.exe", "gevent", "1.0b1"),
    ("gevent-1.0b1.win32-py2.7.msi", "gevent", "1.0b1"),
    ("greenlet-0.3.4-py3.1-win-amd64.egg", "greenlet", "0.3.4"),
    ("greenlet-0.3.4.win-amd64-py3.2.exe", "greenlet", "0.3.4"),
    ("greenlet-0.3.4-py3.2-win32.egg", "greenlet", "0.3.4"),
    ("greenlet-0.3.4-py2.7-linux-x86_64.egg", "greenlet", "0.3.4"),
    ("pep8-0.6.0.zip", "pep8", "0.6.0"),
    ("ABC12-34_V1X-1.2.3.zip", "ABC12", "34_V1X-1.2.3"),
    ("A100-200-XYZ-1.2.3.zip", "A100-200-XYZ", "1.2.3"),
    ("flup-1.0.3.dev-20110405.tar.gz", "flup", "1.0.3.dev-20110405"),
    ("package-1.0.0-alpha.1.zip", "package", "1.0.0-alpha.1"),
    ("package-1.3.7+build.11.e0f985a.zip", "package", "1.3.7+build.11.e0f985a"),
    ("package-v1-8.1.301.ga0df26f.zip", "package-v1", "8.1.301.ga0df26f"),
    ("package-v1.1-8.1.301.ga0df26f.zip", "package-v1.1", "8.1.301.ga0df26f"),
    ("package-2013.02.17.dev123.zip", "package", "2013.02.17.dev123"),
    ("package-20000101.zip", "package", "20000101"),
    ("flup-123-1.0.3.dev-20110405.tar.gz", "flup-123", "1.0.3.dev-20110405"),
    ("package-123-1.0.0-alpha.1.zip", "package-123", "1.0.0-alpha.1"),
    (
        "package-123-1.3.7+build.11.e0f985a.zip",
        "package-123",
        "1.3.7+build.11.e0f985a",
    ),
    ("package-123-v1.1_3-8.1.zip", "package-123-v1.1_3", "8.1"),
    ("package-123-2013.02.17.dev123.zip", "package-123", "2013.02.17.dev123"),
    ("package-123-20000101.zip", "package-123", "20000101"),
    (
        "pyelasticsearch-0.5-brainbot-1-20130712.zip",
        "pyelasticsearch",
        "0.5-brainbot-1-20130712",
    ),
    ("pywin32-217-cp27-none-win32.whl", "pywin32", "217"),
    ("pywin32-217-55-cp27-none-win32.whl", "pywin32", "217-55"),
    ("pywin32-217.1-cp27-none-win32.whl", "pywin32", "217.1"),
    ("package.zip", "package", ""),
    (
        "package-name-0.0.1.dev0.linux-x86_64.tar.gz",
        "package-name",
        "0.0.1.dev0",
    ),
    (
        "package-name-0.0.1.dev0.macosx-10.10-intel.tar.gz",
        "package-name",
        "0.0.1.dev0",
    ),
    (
        "package-name-0.0.1.alpha.1.win-amd64-py3.2.exe",
        "package-name",
        "0.0.1.alpha.1",
    ),
    ("pkg-3!1.0-0.1.tgz", "pkg", "3!1.0-0.1"),  # TO BE FIXED
    ("pkg-3!1+.0-0.1.tgz", "pkg", "3!1+.0-0.1"),  # TO BE FIXED
    ("pkg.zip", "pkg", ""),
    ("foo/pkg.zip", "pkg", ""),
    ("foo/pkg-1b.zip", "pkg", "1b"),
    ("foo/pywin32-217.1-cp27-none-win32.whl", "pywin32", "217.1"),
    (
        "package-name-0.0.1.alpha.1.win-amd64-py3.2.exe",
        "package-name",
        "0.0.1.alpha.1",
    ),
]


def _capitalize_ext(fpath):
    f, e = os.path.splitext(fpath)
    if e != ".whl":
        e = e.upper()
    return f + e


@pytest.mark.parametrize(("filename", "pkgname", "version"), files)
def test_guess_pkgname_and_version(filename, pkgname, version):
    exp = (pkgname, version)
    assert guess_pkgname_and_version(filename) == exp
    assert guess_pkgname_and_version(_capitalize_ext(filename)) == exp


@pytest.mark.parametrize(("filename", "pkgname", "version"), files)
def test_guess_pkgname_and_version_asc(filename, pkgname, version):
    exp = (pkgname, version)
    filename = f"{filename}.asc"
    assert guess_pkgname_and_version(filename) == exp


invalid_files = ["some_file", "some_file.ext", "some_wheel.whl"]


@pytest.mark.parametrize("filename", invalid_files)
def test_guess_pkgname_and_version_invalid_files(filename):
    assert guess_pkgname_and_version(filename) is None


paths = [
    ("/some/path", True),
    (PureWindowsPath(r"c:\some\windows\path"), True),
    ("/.hidden", False),
    (PureWindowsPath(r"c:\.hidden\windows\path"), False),
]


@pytest.mark.parametrize(("pathname", "allowed"), paths)
def test_allowed_path_check(pathname, allowed):
    assert is_listed_path(pathname) == allowed



"""Test doubles."""


class Namespace:
    """Simple namespace."""

    def __init__(self, **kwargs):
        """Instantiate the namespace with the provided kwargs."""
        for k, v in kwargs.items():
            setattr(self, k, v)



import functools
import pathlib
import re as _re
import sys
import typing as t

from pypiserver.bottle import Bottle
from pypiserver.config import Config, RunConfig, strtobool

version = __version__ = "2.2.0"
__version_info__ = tuple(_re.split("[.-]", __version__))
__updated__ = "2024-09-15 21:42:19"

__title__ = "pypiserver"
__summary__ = "A minimal PyPI server for use with pip/easy_install."
__uri__ = "https://github.com/pypiserver/pypiserver"


identity = lambda x: x


def backwards_compat_kwargs(kwargs: dict, warn: bool = True) -> dict:
    """Return a dict with deprecated kwargs converted to new kwargs.

    :param kwargs: the incoming kwargs to convert
    :param warn: whether to output a warning to stderr if there are deprecated
        kwargs found in the incoming kwargs
    """
    # A mapping of deprecated kwargs to a 2-tuple of their corresponding updated
    # kwarg and a function to convert the value of the deprecated kwarg to a
    # value for the new kwarg. `identity` is just a function that returns
    # whatever it is passed and is used in cases where the only change from
    # a legacy kwarg is its name.
    backwards_compat = {
        "authenticated": ("authenticate", identity),
        "passwords": ("password_file", identity),
        # `root` could be a string or an array of strings. Handle both cases,
        # converting strings to Path instances.
        "root": (
            "roots",
            lambda root: [
                # Convert strings to absolute Path instances
                pathlib.Path(r).expanduser().resolve()
                for r in ([root] if isinstance(root, str) else root)
            ],
        ),
        # `redirect_to_fallback` was changed to `disable_fallback` for clearer
        # use as a flag to disable the default behavior. Since its behavior
        # is the opposite, we negate it.
        "redirect_to_fallback": (
            "disable_fallback",
            lambda redirect: not redirect,
        ),
        "server": ("server_method", identity),
        # `welcome_msg` now is just provided as text, so that anyone using
        # pypiserver as a library doesn't need to worry about distributing
        # files if they don't need to. If we're still passed an old-style
        # `welcome_file` argument, we go ahead and resolve it to an absolute
        # path and read the text.
        "welcome_file": (
            "welcome_msg",
            lambda p: pathlib.Path(p).expanduser().resolve().read_text(),
        ),
    }
    # Warn the user if they're using any deprecated arguments
    if warn and any(k in backwards_compat for k in kwargs):
        # Make nice instructions like `Please replace the following:
        # 'authenticated' with 'authenticate'` and print to stderr.
        replacement_strs = (
            f"'{k}' with '{backwards_compat[k][0]}'"
            for k in filter(lambda k: k in kwargs, backwards_compat)
        )
        warn_str = (
            "You are using deprecated arguments. Please replace the following: \n"
            f"  {', '.join(replacement_strs)}"
        )
        print(warn_str, file=sys.stderr)

    # Create an iterable of 2-tuple to collect into the updated dictionary. Each
    # item will either be the existing key-value pair from kwargs, or, if the
    # keyword is a legacy keyword, the new key and potentially adjusted value
    # for that keyword. Note that depending on the order the argument are
    # specified, this _could_ mean an updated legacy keyword could override
    # a new argument if that argument is also specified. However, in that
    # case, our updated kwargs dictionary would have a different number of
    # keys compared to our incoming dictionary, so we check for that case
    # below.
    rv_iter = (
        (
            (k, v)
            if k not in backwards_compat
            else (backwards_compat[k][0], backwards_compat[k][1](v))
        )
        for k, v in kwargs.items()
    )
    updated_kwargs = dict(rv_iter)

    # If our dictionaries have different lengths, we must have gotten duplicate
    # legacy/modern keys. Figure out which keys were dupes and throw an error.
    if len(updated_kwargs) != len(kwargs):
        legacy_to_modern = {k: v[0] for k, v in backwards_compat.items()}
        dupes = [
            (k, v)
            for k, v in legacy_to_modern.items()
            if k in kwargs and v in kwargs
        ]
        raise ValueError(
            "Keyword arguments for pypiserver app() constructor contained "
            "duplicate legacy and modern keys. Duplicates are shown below, in "
            "the form (legacy_key, modern_key):\n"
            f"{dupes}"
        )

    return updated_kwargs


def app(**kwargs: t.Any) -> Bottle:
    """Construct a bottle app running pypiserver.

    :param kwds: Any overrides for defaults. Any property of RunConfig
        (or its base), defined in `pypiserver.config`, may be overridden.
    """
    config = Config.default_with_overrides(**backwards_compat_kwargs(kwargs))
    return setup_routes_from_config(app_from_config(config), config)


def app_from_config(config: RunConfig) -> Bottle:
    """Construct a bottle app from the provided RunConfig."""
    # The _app module instantiates a Bottle instance directly when it is
    # imported. That is `_app.app`. We directly mutate some global variables
    # on the imported `_app` module so that its endpoints will behave as
    # we expect.
    _app = __import__("_app", globals(), locals(), ["."], 1)
    # Because we're about to mutate our import, we pop it out of the imported
    # modules map, so that any future imports do not receive our mutated version
    sys.modules.pop("pypiserver._app", None)
    _app.config = config
    # Add a reference to our config on the Bottle app for easy access in testing
    # and other contexts.
    _app.app._pypiserver_config = config
    return _app.app


def setup_routes_from_config(app: Bottle, config: RunConfig) -> Bottle:
    """Set up additional routes supplied from the config."""

    def _setup_health_endpoint(app, config):
        if config.health_endpoint in [route.rule for route in app.routes]:
            raise RuntimeError(
                "Provided health endpoint overlaps with existing routes"
            )
        app.route(config.health_endpoint, "GET", lambda: "Ok")

    _setup_health_endpoint(app, config)
    return app


T = t.TypeVar("T")


def paste_app_factory(_global_config, **local_conf):
    """Parse a paste config and return an app.

    The paste config is entirely strings, so we need to parse those
    strings into values usable for the config, if they're present.
    """

    def to_bool(val: t.Optional[str]) -> t.Optional[bool]:
        """Convert a string value, if provided, to a bool."""
        return val if val is None else strtobool(val)

    def to_int(val: t.Optional[str]) -> t.Optional[int]:
        """Convert a string value, if provided, to an int."""
        return val if val is None else int(val)

    def to_list(
        val: t.Optional[str],
        sep: str = " ",
        transform: t.Callable[[str], T] = str.strip,
    ) -> t.Optional[t.List[T]]:
        """Convert a string value, if provided, to a list.

        :param sep: the separator between items in the string representation
            of the list
        :param transform: an optional function to call on each string item of
            the list
        """
        if val is None:
            return val
        return list(filter(None, map(transform, val.split(sep))))

    def _make_root(root: str) -> pathlib.Path:
        """Convert a specified string root into an absolute Path instance."""
        return pathlib.Path(root.strip()).expanduser().resolve()

    # A map of config keys we expect in the paste config to the appropriate
    # function to parse the string config value. This map includes both
    # current and legacy keys.
    maps = {
        "cache_control": to_int,
        "roots": functools.partial(to_list, sep="\n", transform=_make_root),
        # root is a deprecated argument for roots
        "root": functools.partial(to_list, sep="\n", transform=_make_root),
        "disable_fallback": to_bool,
        # redirect_to_fallback is a deprecated argument for disable_fallback
        "redirect_to_fallback": to_bool,
        "overwrite": to_bool,
        "authenticate": functools.partial(to_list, sep=" "),
        # authenticated is a deprecated argument for authenticate
        "authenticated": functools.partial(to_list, sep=" "),
        "verbosity": to_int,
    }

    # First, convert values from strings to whatever types we need, or leave
    # them as strings if there's no mapping function available for them.
    mapped_conf = {k: maps.get(k, identity)(v) for k, v in local_conf.items()}
    # Convert any legacy key/value pairs into their modern form.
    updated_conf = backwards_compat_kwargs(mapped_conf)

    return app(**updated_conf)



#! /usr/bin/env python3
"""minimal PyPI like server for use with pip/easy_install"""

import mimetypes
import typing as t
from urllib.parse import quote

from pypiserver.pkg_helpers import normalize_pkgname, parse_version

mimetypes.add_type("application/octet-stream", ".egg")
mimetypes.add_type("application/octet-stream", ".whl")
mimetypes.add_type("text/plain", ".asc")


def get_bad_url_redirect_path(request, project):
    """Get the path for a bad root url."""
    uri = request.custom_fullpath
    if uri.endswith("/"):
        uri = uri[:-1]
    uri = uri.rsplit("/", 1)[0]
    project = quote(project)
    uri += f"/simple/{project}/"
    return uri


class PkgFile:
    __slots__ = [
        "pkgname",  # The projects/package name with possible capitalization
        "version",  # The package version as a string
        "fn",  # The full file path
        "root",  # An optional root directory of the file
        "relfn",  # The file path relative to the root
        "replaces",  # The previous version of the package (used by manage.py)
        "pkgname_norm",  # The PEP503 normalized project name
        "digest",  # The file digest in the form of <algo>=<hash>
        "relfn_unix",  # The relative file path in unix notation
        "parsed_version",  # The package version as a tuple of parts
        "digester",  # a function that calculates the digest for the package
    ]
    digest: t.Optional[str]
    digester: t.Optional[t.Callable[["PkgFile"], t.Optional[str]]]
    parsed_version: tuple
    relfn_unix: t.Optional[str]

    def __init__(
        self,
        pkgname: str,
        version: str,
        fn: t.Optional[str] = None,
        root: t.Optional[str] = None,
        relfn: t.Optional[str] = None,
        replaces: t.Optional["PkgFile"] = None,
    ):
        self.pkgname = pkgname
        self.pkgname_norm = normalize_pkgname(pkgname)
        self.version = version
        self.parsed_version: tuple = parse_version(version)
        self.fn = fn
        self.root = root
        self.relfn = relfn
        self.relfn_unix = None if relfn is None else relfn.replace("\\", "/")
        self.replaces = replaces
        self.digest = None
        self.digester = None

    def __repr__(self) -> str:
        return "{}({})".format(
            self.__class__.__name__,
            ", ".join(
                [
                    f"{k}={getattr(self, k, 'AttributeError')!r}"
                    for k in sorted(self.__slots__)
                ]
            ),
        )

    @property
    def fname_and_hash(self) -> str:
        if self.digest is None and self.digester is not None:
            self.digest = self.digester(self)
        hashpart = f"#{self.digest}" if self.digest else ""
        return self.relfn_unix + hashpart  # type: ignore



#! /usr/bin/env python3
"""Entrypoint for pypiserver."""

import enum
import importlib
import logging
import sys
import typing as t
from pathlib import Path
from wsgiref.simple_server import WSGIRequestHandler

import functools as ft
from pypiserver.config import Config, UpdateConfig

log = logging.getLogger("pypiserver.main")


def init_logging(
    level: int = logging.NOTSET,
    frmt: str = None,
    filename: t.Union[str, Path] = None,
    stream: t.Optional[t.IO] = sys.stderr,
    logger: logging.Logger = None,
) -> None:
    """Configure the specified logger, or the root logger otherwise."""
    logger = logger or logging.getLogger()
    logger.setLevel(level)

    formatter = logging.Formatter(frmt)
    if len(logger.handlers) == 0 and stream is not None:
        handler = logging.StreamHandler(stream)
        handler.setFormatter(formatter)
        logger.addHandler(logging.StreamHandler(stream))

    if filename:
        handler = logging.FileHandler(filename)
        handler.setFormatter(formatter)
        logger.addHandler(handler)


class WsgiHandler(WSGIRequestHandler):
    """A simple request handler to configure logging."""

    # The default `FixedHandler` that bottle's `WSGIRefServer` uses does not
    # log in a particularly predictable or configurable way. We'll pass this
    # in to use instead.
    def address_string(self) -> str:  # Prevent reverse DNS lookups please.
        # This method copied directly from bottle's `FixedHandler` and
        # maintained on the Chesterton's fence principle (i.e. I don't know
        # why it's important, so I'm not going to get rid of it)
        return self.client_address[0]

    def log_message(
        self, format: str, *args: t.Any  # pylint: disable=redefined-builtin
    ) -> None:
        """Log a message."""
        # The log_message method on the `HttpRequestHandler` base class just
        # writes directly to stderr. We'll use its same formatting, but pass
        # it through the logger instead.
        log.info(
            "%s - - [%s] %s\n",
            self.address_string(),
            self.log_date_time_string(),
            format % args,
        )


class AutoServer(enum.Enum):
    """Expected servers that can be automaticlaly selected by bottle."""

    Waitress = enum.auto()
    Paste = enum.auto()
    Twisted = enum.auto()
    CherryPy = enum.auto()
    WsgiRef = enum.auto()


# Possible automatically selected servers. This MUST match the available
# auto servers in bottle.py
AUTO_SERVER_IMPORTS = (
    (AutoServer.Waitress, "waitress"),
    (AutoServer.Paste, "paste"),
    (AutoServer.Twisted, "twisted.web"),
    (AutoServer.CherryPy, "cheroot.wsgi"),
    (AutoServer.CherryPy, "cherrypy.wsgiserver"),
    # this should always be available because it's part of the stdlib
    (AutoServer.WsgiRef, "wsgiref"),
)


def _can_import(name: str) -> bool:
    """Attempt to import a module. Return a bool indicating success."""
    try:
        importlib.import_module(name)
        return True
    except ImportError:
        return False


def guess_auto_server() -> AutoServer:
    """Guess which server bottle will use for the auto setting."""
    # Return the first server that can be imported.
    server = next(
        (s for s, i in AUTO_SERVER_IMPORTS if _can_import(i)),
        None,
    )
    if server is None:
        raise RuntimeError(
            "Unexpected error determining bottle auto server. There may be an "
            "issue with this python environment. Please report this bug at "
            "https://github.com/pypiserver/pypiserver/issues"
        )
    return server


def main(argv: t.Sequence[str] = None) -> None:
    """Application entrypoint for pypiserver.

    This function drives the application (as opposed to the library)
    implementation of pypiserver. Usage from the commandline will result in
    this function being called.
    """
    # pylint: disable=import-outside-toplevel
    import pypiserver  # pylint: disable=redefined-outer-name

    if argv is None:
        # The first item in sys.argv is the name of the python file being
        # executed, which we don't need
        argv = sys.argv[1:]

    config = Config.from_args(argv)

    init_logging(
        level=config.log_level,
        filename=config.log_file,
        frmt=config.log_frmt,
        stream=config.log_stream,
    )

    # Check to see if we were asked to run an update command instead of running
    # the server
    if isinstance(config, UpdateConfig):
        from pypiserver.manage import update_all_packages

        update_all_packages(
            config.roots,
            config.download_directory,
            dry_run=not config.execute,
            stable_only=config.allow_unstable,
            ignorelist=config.ignorelist,
        )
        return

    # Fixes #49:
    #    The gevent server adapter needs to patch some
    #    modules BEFORE importing bottle!
    if config.server_method.startswith("gevent"):
        import gevent.monkey  # @UnresolvedImport

        gevent.monkey.patch_all()

    from pypiserver import bottle

    bottle.debug(config.verbosity > 1)
    bottle._stderr = ft.partial(  # pylint: disable=protected-access
        _logwrite, logging.getLogger(bottle.__name__), logging.INFO
    )

    # Here `app` is a Bottle instance, which we pass to bottle.run() to run
    # the server
    app = pypiserver.app_from_config(config)
    app = pypiserver.setup_routes_from_config(app, config)

    if config.server_method == "gunicorn":
        # When bottle runs gunicorn, gunicorn tries to pull its arguments from
        # sys.argv. Because pypiserver's arguments don't match gunicorn's,
        # this leads to errors.
        # Gunicorn can be configured by using a `gunicorn.conf.py` config file
        # or by specifying the `GUNICORN_CMD_ARGS` env var. See gunicorn
        # docs for more info.
        sys.argv = ["gunicorn"]

    wsgi_kwargs = {"handler_class": WsgiHandler}

    if config.server_method == "auto":
        expected_server = guess_auto_server()
        extra_kwargs = (
            wsgi_kwargs if expected_server is AutoServer.WsgiRef else {}
        )
        log.debug(
            "Server 'auto' selected. Expecting bottle to run '%s'. "
            "Passing extra keyword args: %s",
            expected_server.name,
            extra_kwargs,
        )
    else:
        extra_kwargs = wsgi_kwargs if config.server_method == "wsgiref" else {}
        log.debug(
            "Running bottle with selected server '%s'", config.server_method
        )

    bottle.run(
        app=app,
        host=config.host,
        port=config.port,
        server=config.server_method,
        **extra_kwargs,
    )


def _logwrite(logger, level, msg):
    if msg:
        line_endings = ["\r\n", "\n\r", "\n"]
        for le in line_endings:  # pylint: disable=invalid-name
            if msg.endswith(le):
                msg = msg[: -len(le)]
        if msg:
            logger.log(level, msg)


if __name__ == "__main__":
    main()



import os
import re
import typing as t
from pathlib import PurePath, Path
from urllib.parse import quote


def normalize_pkgname(name: str) -> str:
    """Perform PEP 503 normalization"""
    return re.sub(r"[-_.]+", "-", name).lower()


def normalize_pkgname_for_url(name: str) -> str:
    """Perform PEP 503 normalization and ensure the value is safe for URLs."""
    return quote(normalize_pkgname(name))


# ### Next 2 functions adapted from :mod:`distribute.pkg_resources`.
#


component_re = re.compile(r"(\d+ | [a-z]+ | \.| -)", re.I | re.VERBOSE)
replace = {"pre": "c", "preview": "c", "-": "final-", "rc": "c", "dev": "@"}.get


def _parse_version_parts(s):
    for part in component_re.split(s):
        part = replace(part, part)
        if part in ["", "."]:
            continue
        if part[:1] in "0123456789":
            yield part.zfill(8)  # pad for numeric comparison
        else:
            yield "*" + part

    yield "*final"  # ensure that alpha/beta/candidate are before final


def parse_version(s: str) -> tuple:
    parts = []
    for part in _parse_version_parts(s.lower()):
        if part.startswith("*"):
            # remove trailing zeros from each series of numeric parts
            while parts and parts[-1] == "00000000":
                parts.pop()
        parts.append(part)
    return tuple(parts)


#
# ### -- End of distribute's code.


def is_listed_path(path_part: t.Union[PurePath, str]) -> bool:
    if isinstance(path_part, str):
        path_part = PurePath(path_part)
    return not any(part.startswith(".") for part in path_part.parts)


_archive_suffix_rx = re.compile(
    r"(\.zip|\.tar\.gz|\.tgz|\.tar\.bz2|\.tar\.xz|-py[23]\.\d-.*|"
    r"\.win-amd64-py[23]\.\d\..*|\.win32-py[23]\.\d\..*|\.egg)$",
    re.I,
)
wheel_file_re = re.compile(
    r"""^(?P<namever>(?P<name>.+?)-(?P<ver>\d.*?))
    ((-(?P<build>\d.*?))?-(?P<pyver>.+?)-(?P<abi>.+?)-(?P<plat>.+?)
    \.whl|\.dist-info)$""",
    re.VERBOSE,
)
_pkgname_re = re.compile(r"-\d+[a-z_.!+]", re.I)
_pkgname_parts_re = re.compile(
    r"[\.\-](?=cp\d|py\d|macosx|linux|sunos|solaris|irix|aix|cygwin|win)", re.I
)


def _guess_pkgname_and_version_wheel(
    basename: str,
) -> t.Optional[t.Tuple[str, str]]:
    m = wheel_file_re.match(basename)
    if not m:
        return None
    name = m.group("name")
    ver = m.group("ver")
    build = m.group("build")
    if build:
        return name, ver + "-" + build
    else:
        return name, ver


def guess_pkgname_and_version(path: str) -> t.Optional[t.Tuple[str, str]]:
    path = os.path.basename(path)
    if path.endswith(".asc"):
        path = path.rstrip(".asc")
    if path.endswith(".whl"):
        return _guess_pkgname_and_version_wheel(path)
    if not _archive_suffix_rx.search(path):
        return None
    path = _archive_suffix_rx.sub("", path)
    if "-" not in path:
        pkgname, version = path, ""
    elif path.count("-") == 1:
        pkgname, version = path.split("-", 1)
    elif "." not in path:
        pkgname, version = path.rsplit("-", 1)
    else:
        pkgname = _pkgname_re.split(path)[0]
        ver_spec = path[len(pkgname) + 1 :]
        parts = _pkgname_parts_re.split(ver_spec)
        version = parts[0]
    return pkgname, version



import logging
import os
import pathlib
import sys
import typing as t
from unittest import mock

import pytest

import pypiserver.bottle
from pypiserver import __main__
from pypiserver.bottle import Bottle


THIS_DIR = pathlib.Path(__file__).parent
HTPASS_FILE = THIS_DIR / "../fixtures/htpasswd.a.a"
IGNORELIST_FILE = THIS_DIR / "test-ignorelist"


class main_wrapper:
    app: t.Optional[Bottle]
    run_kwargs: t.Optional[dict]
    update_args: t.Optional[tuple]
    update_kwargs: t.Optional[dict]

    def __init__(self):
        self.app = None
        self.run_kwargs = None
        self.update_args = None
        self.update_kwargs = None

    def __call__(self, argv):
        sys.stdout.write(f"Running {argv}\n")
        # always sets the package directory to this directory, regardless of
        # other passed args.
        __main__.main([str(THIS_DIR)] + argv)
        return self.run_kwargs


@pytest.fixture()
def main(monkeypatch):
    main = main_wrapper()

    def run(**kwargs):
        sys.stdout.write(f"RUN: {kwargs}\n")
        app = kwargs.pop("app")
        main.app = app
        main.run_kwargs = kwargs

    def update(*args, **kwargs):
        main.update_args = args
        main.update_kwargs = kwargs

    monkeypatch.setattr("pypiserver.bottle.run", run)
    monkeypatch.setattr("pypiserver.manage.update_all_packages", update)

    return main


def test_default_pkgdir(main):
    main([])
    assert main.app._pypiserver_config.roots == [THIS_DIR]


def test_noargs(main):
    # Assert we're calling with the default host, port, and server, and
    # assume that we've popped `app` off of the bottle args in our `main`
    # fixture.
    exp_kwargs = {"host": "0.0.0.0", "port": 8080, "server": "auto"}
    actual_kwargs = main([])
    # Only assert our expected are are present. We may pass extra kwargs
    # for particular servers, depending on what is available in the python
    # path.
    assert all(map(lambda k: exp_kwargs[k] == actual_kwargs[k], exp_kwargs))


def test_port(main):
    assert main(["--port=8081"])["port"] == 8081
    assert main(["--port", "8081"])["port"] == 8081
    assert main(["-p", "8081"])["port"] == 8081


def test_server(main):
    assert main(["--server=paste"])["server"] == "paste"
    assert main(["--server", "cherrypy"])["server"] == "cherrypy"


def test_wsgiserver_extra_args_present(monkeypatch, main):
    """The wsgi server gets extra keyword arguments."""
    monkeypatch.setattr(
        __main__,
        "guess_auto_server",
        lambda: __main__.AutoServer.WsgiRef,
    )
    assert main([])["handler_class"] is __main__.WsgiHandler


def test_wsgiserver_extra_kwargs_absent(monkeypatch, main):
    """Other servers don't get wsgiserver args."""
    monkeypatch.setattr(
        __main__,
        "guess_auto_server",
        lambda: __main__.AutoServer.Waitress,
    )
    assert "handler_class" not in main([])


def test_root_multiple(main):
    # Remember we're already setting THIS_DIR as a root in the `main` fixture
    main([str(THIS_DIR.parent)])
    assert main.app._pypiserver_config.roots == [
        THIS_DIR,
        THIS_DIR.parent,
    ]


def test_fallback_url(main):
    main(["--fallback-url", "https://pypi.mirror/simple"])
    assert (
        main.app._pypiserver_config.fallback_url == "https://pypi.mirror/simple"
    )


def test_fallback_url_default(main):
    main([])
    assert (
        main.app._pypiserver_config.fallback_url == "https://pypi.org/simple/"
    )


def test_hash_algo_default(main):
    main([])
    assert main.app._pypiserver_config.hash_algo == "sha256"


def test_hash_algo(main):
    main(["--hash-algo=md5"])
    assert main.app._pypiserver_config.hash_algo == "md5"


def test_hash_algo_off(main):
    main(["--hash-algo=off"])
    assert main.app._pypiserver_config.hash_algo is None
    main(["--hash-algo=0"])
    assert main.app._pypiserver_config.hash_algo is None
    main(["--hash-algo=no"])
    assert main.app._pypiserver_config.hash_algo is None
    main(["--hash-algo=false"])
    assert main.app._pypiserver_config.hash_algo is None


def test_hash_algo_BAD(main):
    with pytest.raises(SystemExit) as excinfo:
        main(["--hash-algo BAD"])


def test_logging(main, tmpdir):
    logfile = tmpdir.mkdir("logs").join("test.log")
    main(["-v", "--log-file", logfile.strpath])
    assert logfile.check(), logfile


def test_logging_verbosity(main):
    main([])
    assert logging.getLogger().level == logging.WARN
    main(["-v"])
    assert logging.getLogger().level == logging.INFO
    main(["-v", "-v"])
    assert logging.getLogger().level == logging.DEBUG
    main(["-v", "-v", "-v"])
    assert logging.getLogger().level == logging.NOTSET


@pytest.mark.parametrize(
    "cli_arg, expected_stream",
    [
        ("stderr", sys.stderr),
        ("stdout", sys.stdout),
        ("none", None),
    ],
)
@mock.patch.object(__main__, "init_logging")
def test_log_to_stdout(init_logging, main, cli_arg, expected_stream):
    main(["--log-stream", cli_arg])
    assert init_logging.call_args[1].get("stream") is expected_stream


@pytest.fixture
def dummy_logger():
    logger = logging.getLogger("test")
    yield logger
    logger.handlers = []


def test_init_logging_with_stream(dummy_logger):
    assert not dummy_logger.handlers

    __main__.init_logging(stream=sys.stdout, logger=dummy_logger)
    assert isinstance(dummy_logger.handlers[0], logging.StreamHandler)
    assert dummy_logger.handlers[0].stream is sys.stdout


def test_init_logging_with_none_stream_doesnt_add_stream_handler(dummy_logger):
    assert not dummy_logger.handlers

    __main__.init_logging(stream=None, logger=dummy_logger)
    assert not dummy_logger.handlers


def test_welcome_file(main):
    sample_msg_file = os.path.join(os.path.dirname(__file__), "sample_msg.html")
    main(["--welcome", sample_msg_file])
    assert "Hello pypiserver tester!" in main.app._pypiserver_config.welcome_msg


def test_welcome_file_default(main):
    main([])
    assert "Welcome to pypiserver!" in main.app._pypiserver_config.welcome_msg


def test_password_without_auth_list(main, monkeypatch):
    sysexit = mock.MagicMock(side_effect=ValueError("BINGO"))
    monkeypatch.setattr("sys.exit", sysexit)
    with pytest.raises(ValueError) as ex:
        main(["-P", str(HTPASS_FILE), "-a", ""])
    assert ex.value.args[0] == "BINGO"

    with pytest.raises(ValueError) as ex:
        main(["-a", "."])
    assert ex.value.args[0] == "BINGO"
    with pytest.raises(ValueError) as ex:
        main(["-a", ""])
    assert ex.value.args[0] == "BINGO"

    with pytest.raises(ValueError) as ex:
        main(["-P", "."])
    assert ex.value.args[0] == "BINGO"


def test_password_alone(main, monkeypatch):
    monkeypatch.setitem(sys.modules, "passlib", mock.MagicMock())
    monkeypatch.setitem(sys.modules, "passlib.apache", mock.MagicMock())
    main(["-P", str(HTPASS_FILE)])
    assert main.app._pypiserver_config.authenticate == ["update"]


def test_dot_password_without_auth_list(main, monkeypatch):
    main(["-P", ".", "-a", "."])
    assert main.app._pypiserver_config.authenticate == []


def test_blacklist_file(main):
    """
    Test that calling the app with the --blacklist-file argument does not
    throw a getopt error
    """
    main(["-U", "--blacklist-file", str(IGNORELIST_FILE)])
    assert main.update_kwargs["ignorelist"] == ["mypiserver", "something"]


def test_auto_servers() -> None:
    """Test auto servers."""
    # A list of bottle ServerAdapters
    bottle_adapters = tuple(
        a.__name__.lower() for a in pypiserver.bottle.AutoServer.adapters
    )
    # We are going to expect that our AutoServer enum names must match those
    # at least closely enough to be recognizable.
    our_mappings = tuple(map(str.lower, __main__.AutoServer.__members__))

    # Assert that all of our mappings are represented in bottle adapters
    assert all(
        any(mapping in a for a in bottle_adapters) for mapping in our_mappings
    )

    # Assert that our import checking order matches the order in which the
    # adapters are defined in the AutoServer
    our_check_order = tuple(i[0] for i in __main__.AUTO_SERVER_IMPORTS)

    # Some of the servers have more than one check, so we need to remove
    # duplicates before we check for identity with the AutoServer definition.
    seen: t.Dict[__main__.AutoServer, __main__.AutoServer] = {}
    our_check_order = tuple(
        seen.setdefault(i, i) for i in our_check_order if i not in seen
    )

    # We should have the same number of deduped checkers as there are bottle
    # adapters
    assert len(our_check_order) == len(bottle_adapters)

    # And the order should be the same
    assert all(
        us.name.lower() in them
        for us, them in zip(our_check_order, bottle_adapters)
    )


def test_health_endpoint_default(main):
    main([])
    assert main.app._pypiserver_config.health_endpoint == "/health"
    assert "/health" in (route.rule for route in main.app.routes)


def test_health_endpoint_customized(main):
    main(["--health-endpoint", "/healthz"])
    assert main.app._pypiserver_config.health_endpoint == "/healthz"
    assert "/healthz" in (route.rule for route in main.app.routes)


def test_health_endpoint_invalid_customized(main):
    with pytest.raises(SystemExit):
        main(["--health-endpoint", "/health!"])



"""Pypiserver configuration management.

NOTE: THIS CONFIG IS NOT YET IN USE. It is the intended replacement for
the current config logic, but has not yet been integrated.

To add a config option:

- If it should be available for all subcommands (run, update), add it to
  the `add_common_args()` function
- If it should only be available for the `run` command, add it to the
  `run_parser` in the `get_parser()` function.
- If it should only be available for the `update` command, add it to the
  `update_parser` in the `get_parser() function`.
- Add it to the appropriate Config class, `_ConfigCommon` for global options,
  `RunConfig` for `run` options, and `UpdateConfig` for `update` options.
  - This requires adding it as an `__init__()` kwarg, setting it as an instance
    attribute in `__init__()`, and ensuring it will be parsed from the argparse
    namespace in the `kwargs_from_namespace()` method
- Ensure your config option is tested in `tests/test_config.py`.

The `Config` class is a factory class only. Config objects do not inherit from
it, but from `_ConfigCommon`. The `Config` provides the following constructors:

- `default_with_overrides(**overrides: Any)`: construct a `RunConfig` (since
  run is the default pypiserver action) with default values, applying any
  specified overrides
- `from_args(args: Optional[Sequence[str]])`: construct a config from the
  provided arguments. Depending on arguments, the config will be either a
  `RunConfig` or an `UpdateConfig`

Legacy commandline arguments did not require a subcommand. This form is
still supported, but deprecated. A warning is printing to stderr if
the legacy commandline format is used.
"""

import argparse
import contextlib
import hashlib
import io
import logging
import pathlib
import re
import sys
import textwrap
import typing as t

try:
    # `importlib_resources` is required for Python versions below 3.12
    # See more in the package docs: https://pypi.org/project/importlib-resources/
    try:
        from importlib_resources import files as import_files
    except ImportError:
        from importlib.resources import files as import_files

    def get_resource_bytes(package: str, resource: str) -> bytes:
        ref = import_files(package).joinpath(resource)
        return ref.read_bytes()

except ImportError:
    # The `pkg_resources` is deprecated in Python 3.12
    import pkg_resources

    def get_resource_bytes(package: str, resource: str) -> bytes:
        return pkg_resources.resource_string(package, resource)


from pypiserver.backend import (
    SimpleFileBackend,
    CachingFileBackend,
    Backend,
    IBackend,
    get_file_backend,
    BackendProxy,
)

# The `passlib` requirement is optional, so we need to verify its import here.
try:
    from passlib.apache import HtpasswdFile
except ImportError:
    HtpasswdFile = None


def legacy_strtoint(val: str) -> int:
    """Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.

    The "strtobool" function in distutils does a nice job at parsing strings,
    but returns an integer. This just wraps it in a boolean call so that we
    get a bool.

    Borrowed from deprecated distutils.
    """
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return 1
    elif val in ("n", "no", "f", "false", "off", "0"):
        return 0
    else:
        raise ValueError("invalid truth value {!r}".format(val))


strtobool: t.Callable[[str], bool] = lambda val: bool(legacy_strtoint(val))


# Specify defaults here so that we can use them in tests &c. and not need
# to update things in multiple places if a default changes.
class DEFAULTS:
    """Config defaults."""

    AUTHENTICATE = ["update"]
    FALLBACK_URL = "https://pypi.org/simple/"
    HEALTH_ENDPOINT = "/health"
    HASH_ALGO = "sha256"
    INTERFACE = "0.0.0.0"
    LOG_FRMT = "%(asctime)s|%(name)s|%(levelname)s|%(thread)d|%(message)s"
    LOG_ERR_FRMT = "%(body)s: %(exception)s \n%(traceback)s"
    LOG_REQ_FRMT = "%(bottle.request)s"
    LOG_RES_FRMT = "%(status)s"
    LOG_STREAM = sys.stdout
    PACKAGE_DIRECTORIES = [pathlib.Path("~/packages").expanduser().resolve()]
    PORT = 8080
    SERVER_METHOD = "auto"
    BACKEND = "auto"


def auth_arg(arg: str) -> t.List[str]:
    """Parse the authentication argument."""
    # Split on commas, remove duplicates, remove whitespace, ensure lowercase.
    # Sort so that they'll have a consistent ordering.
    items = sorted(list(set(i.strip().lower() for i in arg.split(","))))
    # Throw for any invalid options
    if any(i not in ("download", "list", "update", ".") for i in items):
        raise ValueError(
            "Invalid authentication option. Valid values are download, list, "
            "and update, or . (for no authentication)."
        )
    # The "no authentication" option must be specified in isolation.
    if "." in items and len(items) > 1:
        raise argparse.ArgumentTypeError(
            "Invalid authentication options. `.` (no authentication) "
            "must be specified alone."
        )

    # The "." is just an indicator for no auth, so we return an empty auth list
    # if it was present.
    return [i for i in items if not i == "."]


def hash_algo_arg(arg: str) -> t.Optional[str]:
    """Parse a hash algorithm from the string."""
    if arg in hashlib.algorithms_available:
        return arg
    try:
        if not strtobool(arg):
            return None
    except ValueError:
        # strtobool raises if the string doesn't seem like a truthiness-
        # indicating string. We do want to raise in this case, but we want
        # to raise our own custom message rather than raising the ValueError
        # raised by strtobool.
        pass
    # At this point we either had an invalid hash algorithm or a truthy string
    # like 'yes' or 'true'. In either case, we want to throw an error.
    raise argparse.ArgumentTypeError(
        f"Hash algorithm '{arg}' is not available. Please select one "
        f"of {hashlib.algorithms_available}, or turn off hashing by "
        "setting --hash-algo to 'off', '0', or 'false'"
    )


def health_endpoint_arg(arg: str) -> str:
    """Verify the health_endpoint and raises ValueError if invalid."""
    rule_regex = r"^/[a-z0-9/_-]+$"
    if re.fullmatch(rule_regex, arg, re.I) is not None:
        return arg

    raise argparse.ArgumentTypeError(
        "Invalid path for the health endpoint. Make sure that it contains only "
        "alphanumeric characters, hyphens, forward slashes and underscores. "
        f"In other words, make sure to match the following regex: {rule_regex}"
    )


def html_file_arg(arg: t.Optional[str]) -> str:
    """Parse the provided HTML file and return its contents."""
    if arg is None or arg == "pypiserver/welcome.html":
        return get_resource_bytes(__name__, "welcome.html").decode("utf-8")
    with open(arg, "r", encoding="utf-8") as f:
        msg = f.read()
    return msg


def ignorelist_file_arg(arg: t.Optional[str]) -> t.List[str]:
    """Parse the ignorelist and return the list of ignored files."""
    if arg is None or arg == "pypiserver/no-ignores":
        return []

    fpath = pathlib.Path(arg)
    if not fpath.exists():
        raise argparse.ArgumentTypeError(f"No such ignorelist-file '{arg}'")

    try:
        lines = (ln.strip() for ln in fpath.read_text().splitlines())
        return [ln for ln in lines if ln and not ln.startswith("#")]
    except Exception as exc:
        raise argparse.ArgumentTypeError(
            f"Could not parse ignorelist-file '{arg}': {exc}"
        ) from exc


def package_directory_arg(arg: str) -> pathlib.Path:
    """Convert any package directory argument into its absolute path."""
    pkg_dir = pathlib.Path(arg).expanduser().resolve()
    try:
        # Attempt to grab the first item from the directory. The directory may
        # be empty, in which case we'll get back None, but if the directory does
        # not exist or we do not have permission to read it, we can catch th
        # OSError and exit with a useful message.
        next(pkg_dir.iterdir(), None)
    except OSError as exc:
        raise argparse.ArgumentTypeError(
            "Error: while trying to access package directory "
            f"({pkg_dir}): {exc}"
        )
    return pkg_dir


# We need to capture this at compile time, because we replace sys.stderr
# during config parsing in order to better control error output when we
# encounter legacy cmdline arguments.
_ORIG_STDERR = sys.stderr


def log_stream_arg(arg: str) -> t.Optional[t.IO]:
    """Parse the log-stream argument."""
    lower = arg.lower()
    # Convert a `none` string to a real none.
    val = lower if lower != "none" else None
    # Ensure the remaining value is a valid stream type, and return it.
    if val is None:
        return val
    if val == "stdout":
        return sys.stdout
    if val == "stderr":
        return _ORIG_STDERR
    raise argparse.ArgumentTypeError(
        "Invalid option for --log-stream. Value must be one of stdout, "
        "stderr, or none."
    )


def add_common_args(parser: argparse.ArgumentParser) -> None:
    """Add common arguments to a parser."""
    # Don't update at top-level to avoid circular imports in __init__
    from pypiserver import __version__

    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Enable verbose logging; repeat for more verbosity.",
    )
    parser.add_argument(
        "--log-file",
        metavar="FILE",
        help=(
            "Write logging info into this FILE, as well as to stdout or "
            "stderr, if configured."
        ),
    )
    parser.add_argument(
        "--log-stream",
        metavar="STREAM",
        default=DEFAULTS.LOG_STREAM,
        type=log_stream_arg,
        help=(
            "Log messages to the specified STREAM. Valid values are stdout, "
            "stderr, and none"
        ),
    )
    parser.add_argument(
        "--log-frmt",
        metavar="FORMAT",
        default=DEFAULTS.LOG_FRMT,
        help=(
            "The logging format-string.  (see `logging.LogRecord` class from "
            "standard python library)"
        ),
    )

    parser.add_argument(
        "--hash-algo",
        default=DEFAULTS.HASH_ALGO,
        type=hash_algo_arg,
        help=(
            "Any `hashlib` available algorithm to use for generating fragments "
            "on package links. Can be disabled with one of (0, no, off, false)."
        ),
    )

    parser.add_argument(
        "--backend",
        default=DEFAULTS.BACKEND,
        choices=("auto", "simple-dir", "cached-dir"),
        dest="backend_arg",
        help=(
            "A backend implementation. Keep the default 'auto' to automatically"
            " determine whether to activate caching or not"
        ),
    )

    parser.add_argument(
        "--version",
        action="version",
        version=__version__,
    )


def get_parser() -> argparse.ArgumentParser:
    """Return an ArgumentParser."""
    parser = argparse.ArgumentParser(
        description=(
            "start PyPI compatible package server serving packages from "
            "PACKAGES_DIRECTORY. If PACKAGES_DIRECTORY is not given on the "
            "command line, it uses the default ~/packages. pypiserver scans "
            "this directory recursively for packages. It skips packages and "
            "directories starting with a dot. Multiple package directories "
            "may be specified."
        ),
        formatter_class=PreserveWhitespaceRawTextHelpFormatter,
        epilog=(
            "Visit https://github.com/pypiserver/pypiserver "
            "for more information\n \n"
        ),
    )

    add_common_args(parser)

    subparsers = parser.add_subparsers(dest="cmd")

    run_parser = subparsers.add_parser(
        "run",
        formatter_class=PreserveWhitespaceRawTextHelpFormatter,
        help="Run pypiserver, serving packages from PACKAGES_DIRECTORY",
    )

    add_common_args(run_parser)

    run_parser.add_argument(
        "package_directory",
        default=DEFAULTS.PACKAGE_DIRECTORIES,
        nargs="*",
        type=package_directory_arg,
        help="The directory from which to serve packages.",
    )

    run_parser.add_argument(
        "-p",
        "--port",
        type=int,
        default=DEFAULTS.PORT,
        help="Listen on port PORT (default: 8080)",
    )
    run_parser.add_argument(
        "-i",
        "-H",
        "--interface",
        "--host",
        dest="host",
        default=DEFAULTS.INTERFACE,
        help="Listen on interface INTERFACE (default: 0.0.0.0)",
    )
    run_parser.add_argument(
        "-a",
        "--authenticate",
        default=DEFAULTS.AUTHENTICATE,
        type=auth_arg,
        help=(
            "Comma-separated list of (case-insensitive) actions to "
            "authenticate (options: download, list, update; default: update)."
            "\n\n "
            "Any actions not specified are not authenticated, so to "
            "authenticate downloads and updates, but allow unauthenticated "
            "viewing of the package list, you would use: "
            "\n\n"
            "  pypi-server -a 'download, update' -P ./my_passwords.htaccess"
            "\n\n"
            "To disable authentication, use:"
            "\n\n"
            "  pypi-server -a . -P ."
            "\n\n"
            "See the `-P` option for configuring users and passwords. "
            "\n\n"
            "Note that when uploads are not protected, the `register` command "
            "is not necessary, but `~/.pypirc` still needs username and "
            "password fields, even if bogus."
        ),
    )
    run_parser.add_argument(
        "-P",
        "--passwords",
        metavar="PASSWORD_FILE",
        help=(
            "Use an apache htpasswd file PASSWORD_FILE to set usernames and "
            "passwords for authentication."
            "\n\n"
            "To allow unauthorized access, use:"
            "\n\n"
            "  pypi-server -a . -P ."
            "\n\n"
        ),
    )
    run_parser.add_argument(
        "--disable-fallback",
        action="store_true",
        help=(
            "Disable the default redirect to PyPI for packages not found in "
            "the local index."
        ),
    )
    run_parser.add_argument(
        "--fallback-url",
        default=DEFAULTS.FALLBACK_URL,
        help=(
            "Redirect to FALLBACK_URL for packages not found in the local "
            "index."
        ),
    )
    run_parser.add_argument(
        "--health-endpoint",
        default=DEFAULTS.HEALTH_ENDPOINT,
        type=health_endpoint_arg,
        help=(
            "Configure a custom liveness endpoint. It always returns 200 Ok if "
            "the service is up. Otherwise, it means that the service is not responsive."
        ),
    )
    run_parser.add_argument(
        "--server",
        metavar="METHOD",
        default=DEFAULTS.SERVER_METHOD,
        choices=(
            "auto",
            "cherrypy",
            "gevent",
            "gunicorn",
            "paste",
            "twisted",
            "wsgiref",
        ),
        type=str.lower,
        help=(
            "Use METHOD to run the server. Valid values include paste, "
            "cherrypy, twisted, gunicorn, gevent, wsgiref, and auto. The "
            'default is to use "auto", which chooses one of paste, cherrypy, '
            "twisted, or wsgiref."
        ),
    )
    run_parser.add_argument(
        "-o",
        "--overwrite",
        action="store_true",
        help="Allow overwriting existing package files during upload.",
    )
    run_parser.add_argument(
        "--welcome",
        metavar="HTML_FILE",
        # we want to run our `html_file_arg` function to get our default value
        # if the value isn't provided, but if we specify `None` as a default
        # or let argparse handle the default logic, it will not call that
        # function with the None value. So, we set it to a custom value.
        default="pypiserver/welcome.html",
        type=html_file_arg,
        help=(
            "Use the contents of HTML_FILE as a custom welcome message "
            "on the home page."
        ),
    )
    run_parser.add_argument(
        "--cache-control",
        metavar="AGE",
        type=int,
        help=(
            'Add "Cache-Control: max-age=AGE" header to package downloads. '
            "Pip 6+ requires this for caching."
            "AGE is specified in seconds."
        ),
    )
    run_parser.add_argument(
        "--log-req-frmt",
        metavar="FORMAT",
        default=DEFAULTS.LOG_REQ_FRMT,
        help=(
            "A format-string selecting Http-Request properties to log; set "
            "to '%%s' to see them all."
        ),
    )
    run_parser.add_argument(
        "--log-res-frmt",
        metavar="FORMAT",
        default=DEFAULTS.LOG_RES_FRMT,
        help=(
            "A format-string selecting Http-Response properties to log; set "
            "to '%%s' to see them all."
        ),
    )
    run_parser.add_argument(
        "--log-err-frmt",
        metavar="FORMAT",
        default=DEFAULTS.LOG_ERR_FRMT,
        help=(
            "A format-string selecting Http-Error properties to log; set "
            "to '%%s' to see them all."
        ),
    )

    update_parser = subparsers.add_parser(
        "update",
        help=textwrap.dedent(
            "Handle updates of packages managed by pypiserver. By default, "
            "a pip command to update the packages is printed to stdout for "
            "introspection or pipelining. See the `-x` option for updating "
            "packages directly."
        ),
    )

    add_common_args(update_parser)

    update_parser.add_argument(
        "package_directory",
        default=DEFAULTS.PACKAGE_DIRECTORIES,
        nargs="*",
        type=package_directory_arg,
        help="The directory from which to serve packages.",
    )

    update_parser.add_argument(
        "-x",
        "--execute",
        action="store_true",
        help="Execute the pip commands rather than printing to stdout",
    )
    update_parser.add_argument(
        "-d",
        "--download-directory",
        help=(
            "Specify a directory where packages updates will be downloaded. "
            "The default behavior is to use the directory which contains "
            "the package being updated."
        ),
    )
    update_parser.add_argument(
        "-u",
        "--allow-unstable",
        action="store_true",
        help=(
            "Allow updating to unstable versions (alpha, beta, rc, dev, etc.)"
        ),
    )
    update_parser.add_argument(
        "--blacklist-file",
        "--ignorelist-file",
        dest="ignorelist_file",
        default="pypiserver/no-ignores",
        type=ignorelist_file_arg,
        help=(
            "Don't update packages listed in this file (one package name per "
            "line, without versions, '#' comments honored). This can be useful "
            "if you upload private packages into pypiserver, but also keep a "
            "mirror of public packages that you regularly update. Attempting "
            "to pull an update of a private package from `pypi.org` might pose "
            "a security risk - e.g. a malicious user might publish a higher "
            "version of the private package, containing arbitrary code."
        ),
    )
    return parser


TConf = t.TypeVar("TConf", bound="_ConfigCommon")
BackendFactory = t.Callable[["_ConfigCommon"], Backend]


class _ConfigCommon:
    hash_algo: t.Optional[str] = None

    def __init__(
        self,
        roots: t.List[pathlib.Path],
        verbosity: int,
        log_frmt: str,
        log_file: t.Optional[str],
        log_stream: t.Optional[t.IO],
        hash_algo: t.Optional[str],
        backend_arg: str,
    ) -> None:
        """Construct a RuntimeConfig."""
        # Global arguments
        self.verbosity = verbosity
        self.log_file = log_file
        self.log_stream = log_stream
        self.log_frmt = log_frmt

        self.roots = roots
        self.hash_algo = hash_algo
        self.backend_arg = backend_arg

        # Derived properties are directly based on other properties and are not
        # included in equality checks.
        self._derived_properties: t.Tuple[str, ...] = (
            "iter_packages",
            "package_root",
            "backend",
        )
        # The first package directory is considered the root. This is used
        # for uploads.
        self.package_root = self.roots[0]

        self.backend = self.get_backend(backend_arg)

    @classmethod
    def from_namespace(
        cls: t.Type[TConf], namespace: argparse.Namespace
    ) -> TConf:
        """Construct a config from an argparse namespace."""
        return cls(**cls.kwargs_from_namespace(namespace))

    @staticmethod
    def kwargs_from_namespace(
        namespace: argparse.Namespace,
    ) -> t.Dict[str, t.Any]:
        """Convert a namespace into __init__ kwargs for this class."""
        return dict(
            verbosity=namespace.verbose,
            log_file=namespace.log_file,
            log_stream=namespace.log_stream,
            log_frmt=namespace.log_frmt,
            roots=namespace.package_directory,
            hash_algo=namespace.hash_algo,
            backend_arg=namespace.backend_arg,
        )

    @property
    def log_level(self) -> int:
        """Return an appropriate log-level for the config's verbosity."""
        levels = {
            0: logging.WARNING,
            1: logging.INFO,
            2: logging.DEBUG,
        }
        # Return a log-level from warning through not set (log all messages).
        # If we've specified 3 or more levels of verbosity, just return not set.
        return levels.get(self.verbosity, logging.NOTSET)

    def get_backend(self, arg: str) -> IBackend:
        available_backends: t.Dict[str, BackendFactory] = {
            "auto": get_file_backend,
            "simple-dir": SimpleFileBackend,
            "cached-dir": CachingFileBackend,
        }

        backend = available_backends[arg]

        return BackendProxy(backend(self))

    def with_updates(self: TConf, **kwargs: t.Any) -> TConf:
        """Create a new config with the specified updates.

        The current config is used as a base. Any properties not specified in
        keyword arguments will remain unchanged.
        """
        return self.__class__(**{**dict(self), **kwargs})

    def __repr__(self) -> str:
        """A string representation indicating the class and its properties."""
        return "{}({})".format(
            self.__class__.__name__,
            ", ".join(
                f"{k}={v}"
                for k, v in vars(self).items()
                if not k.startswith("_")
            ),
        )

    def __eq__(self, other: t.Any) -> bool:
        """Configs are equal if their public values are equal."""
        if not isinstance(other, self.__class__):
            return False
        return all(
            getattr(other, k) == v
            for k, v in self
            if k not in self._derived_properties
        )

    def __iter__(self) -> t.Iterator[t.Tuple[str, t.Any]]:
        """Iterate over config (k, v) pairs."""
        yield from (
            (k, v)
            for k, v in vars(self).items()
            if not k.startswith("_") and k not in self._derived_properties
        )


class RunConfig(_ConfigCommon):
    """A config for the Run command."""

    def __init__(
        self,
        port: int,
        host: str,
        authenticate: t.List[str],
        password_file: t.Optional[str],
        disable_fallback: bool,
        fallback_url: str,
        health_endpoint: str,
        server_method: str,
        overwrite: bool,
        welcome_msg: str,
        cache_control: t.Optional[int],
        log_req_frmt: str,
        log_res_frmt: str,
        log_err_frmt: str,
        auther: t.Optional[t.Callable[[str, str], bool]] = None,
        **kwargs: t.Any,
    ) -> None:
        """Construct a RuntimeConfig."""
        super().__init__(**kwargs)
        self.port = port
        self.host = host
        self.authenticate = authenticate
        self.password_file = password_file
        self.disable_fallback = disable_fallback
        self.fallback_url = fallback_url
        self.health_endpoint = health_endpoint
        self.server_method = server_method
        self.overwrite = overwrite
        self.welcome_msg = welcome_msg
        self.cache_control = cache_control
        self.log_req_frmt = log_req_frmt
        self.log_res_frmt = log_res_frmt
        self.log_err_frmt = log_err_frmt
        # Derived properties
        self._derived_properties = self._derived_properties + ("auther",)
        self.auther = self.get_auther(auther)

    @classmethod
    def kwargs_from_namespace(
        cls, namespace: argparse.Namespace
    ) -> t.Dict[str, t.Any]:
        """Convert a namespace into __init__ kwargs for this class."""
        return {
            **super(RunConfig, cls).kwargs_from_namespace(namespace),
            "port": namespace.port,
            "host": namespace.host,
            "authenticate": namespace.authenticate,
            "password_file": namespace.passwords,
            "disable_fallback": namespace.disable_fallback,
            "fallback_url": namespace.fallback_url,
            "health_endpoint": namespace.health_endpoint,
            "server_method": namespace.server,
            "overwrite": namespace.overwrite,
            "welcome_msg": namespace.welcome,
            "cache_control": namespace.cache_control,
            "log_req_frmt": namespace.log_req_frmt,
            "log_res_frmt": namespace.log_res_frmt,
            "log_err_frmt": namespace.log_err_frmt,
        }

    def get_auther(
        self, passed_auther: t.Optional[t.Callable[[str, str], bool]]
    ) -> t.Callable[[str, str], bool]:
        """Create or retrieve an authentication function."""
        # The auther may be specified directly as a kwarg in the API interface
        if passed_auther:
            return passed_auther
        # Otherwise we check to see if we need to authenticate
        if self.password_file == "." or self.authenticate == []:
            # It's illegal to specify no authentication without also specifying
            # no password file, and vice-versa.
            if self.password_file != "." or self.authenticate != []:
                sys.exit(
                    "When auth-ops-list is empty (-a=.), password-file"
                    f" (-P={self.password_file!r}) must also be empty ('.')!"
                )
            # Return an auther that always returns true.
            return lambda _uname, _pw: True
        # Now, if there was no password file specified, we can return an auther
        # that always returns False, since there is no way to authenticate.
        if self.password_file is None:
            return lambda _uname, _pw: False
        # Finally, if a password file was specified, we'll load it up with
        # Htpasswd and return a callable that checks it.
        if HtpasswdFile is None:
            sys.exit(
                "apache.passlib library is not available. You must install "
                "pypiserver with the optional 'passlib' dependency (`pip "
                "install pypiserver['passlib']`) in order to use password "
                "authentication"
            )

        loaded_pw_file = HtpasswdFile(self.password_file)

        # Construct a local closure over the loaded PW file and return as our
        # authentication function.
        def auther(uname: str, pw: str) -> bool:
            loaded_pw_file.load_if_changed()
            return loaded_pw_file.check_password(uname, pw)

        return auther


class UpdateConfig(_ConfigCommon):
    """A config for the Update command."""

    def __init__(
        self,
        execute: bool,
        download_directory: t.Optional[str],
        allow_unstable: bool,
        ignorelist: t.List[str],
        **kwargs: t.Any,
    ) -> None:
        """Construct an UpdateConfig."""
        super().__init__(**kwargs)
        self.execute = execute
        self.download_directory = download_directory
        self.allow_unstable = allow_unstable
        self.ignorelist = ignorelist

    @classmethod
    def kwargs_from_namespace(
        cls, namespace: argparse.Namespace
    ) -> t.Dict[str, t.Any]:
        """Convert a namespace into __init__ kwargs for this class."""
        return {
            **super(UpdateConfig, cls).kwargs_from_namespace(namespace),
            "execute": namespace.execute,
            "download_directory": namespace.download_directory,
            "allow_unstable": namespace.allow_unstable,
            "ignorelist": namespace.ignorelist_file,
        }


Configuration = t.Union[RunConfig, UpdateConfig]


class Config:
    """Config constructor for building a config from args."""

    @classmethod
    def default_with_overrides(cls, **overrides: t.Any) -> RunConfig:
        """Construct a RunConfig with default arguments, plus overrides.

        Overrides must be valid arguments to the `__init__()` function
        of `RunConfig`.
        """
        default_config = cls.from_args(["run"])
        assert isinstance(default_config, RunConfig)
        return default_config.with_updates(**overrides)

    @classmethod
    def from_args(
        cls, args: t.Optional[t.Sequence[str]] = None
    ) -> Configuration:
        """Construct a Config from the passed args or sys.argv."""
        # If pulling args from sys.argv (commandline arguments), argv[0] will
        # be the program name, (i.e. pypi-server), so we don't need to
        # worry about it.
        args = args if args is not None else sys.argv[1:]
        parser = get_parser()

        try:
            with capture_stderr() as cap:
                parsed = parser.parse_args(args)
            # There's a special case we need to handle where no arguments
            # whatsoever were provided. Because we need to introspect
            # what subcommand is called, via the `add_subparsers(dest='cmd')`
            # call, calling with no subparser is _not_ an error. We will
            # treat it as such, so that we then trigger the legacy argument
            # handling logic.
            if parsed.cmd is None:
                sys.exit(1)
        except SystemExit as exc:
            # A SystemExit is raised in some non-error cases, like
            # printing the help or the version. Reraise in those cases.
            cap.seek(0)
            first_txt = cap.read()
            if not exc.code or exc.code == 0:
                # There usually won't have been any error text in these
                # cases, but if there was, print it.
                if first_txt:
                    print(first_txt, file=sys.stderr)
                raise
            # Otherwise, it's possible they're using the older, non-subcommand
            # form of the CLI. In this case, attempt to parse with adjusted
            # arguments. If the parse is successful, warn them about using
            # deprecated arguments and continue. If this parse _also_ fails,
            # show them the parsing error for their original arguments,
            # not for the adjusted arguments.
            try:
                with capture_stderr() as cap:
                    parsed = parser.parse_args(cls._adjust_old_args(args))
                print(
                    "WARNING: You are using deprecated arguments to pypiserver.\n\n"
                    "Please run `pypi-server --help` and update your command "
                    "to align with the current interface.\n\n"
                    "In most cases, this will be as simple as just using\n\n"
                    "  pypi-server run [args]\n\n"
                    "instead of\n\n"
                    "  pypi-server [args]\n",
                    file=sys.stderr,
                )
            except SystemExit:
                cap.seek(0)
                second_txt = cap.read()
                if not exc.code or exc.code == 0:
                    # Again, usually nothing to stderr in these cases,
                    # but if there was, print it and re-raise.
                    if second_txt:
                        print(second_txt, file=sys.stderr)
                    raise
                # Otherwise, we print the original error text instead of
                # the error text from the call with our adjusted args,
                # and then raise. Showing the original error text will
                # provide a usage error for the new argument form, which
                # should help folks to upgrade.
                if first_txt:
                    print(first_txt, file=sys.stderr)
                raise

        if parsed.cmd == "run":
            return RunConfig.from_namespace(parsed)
        if parsed.cmd == "update":
            return UpdateConfig.from_namespace(parsed)
        raise SystemExit(parser.format_usage())

    @staticmethod
    def _adjust_old_args(args: t.Sequence[str]) -> t.List[str]:
        """Adjust args for backwards compatibility.

        Should only be called once args have been verified to be unparsable.
        """
        # Backwards compatibility hack: for most of pypiserver's life, "run"
        # and "update" were not separate subcommands. The `-U` flag being
        # present on the cmdline, regardless of other arguments, would lead
        # to update behavior. In order to allow subcommands without
        # breaking backwards compatibility, we need to insert "run" or
        # "update" as a positional arg before any other arguments.

        # We will be adding our subcommand as the first argument.
        insertion_idx = 0

        # Don't actually update the passed list, in case it was the global
        # sys.argv.
        updated_args = list(args)

        # Find the update index. For "reasons", python's index search throws
        # if the value isn't found, so manually wrap in the usual "return -1
        # if not found" standard
        try:
            update_idx = updated_args.index("-U")
        except ValueError:
            update_idx = -1

        if update_idx == -1:
            # We were a "run" command.
            updated_args.insert(insertion_idx, "run")
        else:
            # Remove the -U from the args and add the "update" command at the
            # start of the arg list.
            updated_args.pop(update_idx)
            updated_args.insert(insertion_idx, "update")

        return updated_args


@contextlib.contextmanager
def capture_stderr() -> t.Iterator[t.IO]:
    """Capture stderr and yield as a buffer."""
    orig = sys.stderr
    cap = io.StringIO()
    sys.stderr = cap
    try:
        yield cap
    finally:
        sys.stderr = orig


# Note: this is adapted from this StackOverflow answer:
# https://stackoverflow.com/a/35925919 -- the normal "raw" help
# text formatters provided with the argparse library don't do
# a great job of maintaining whitespace while still keeping
# subsequent lines properly intended.
class PreserveWhitespaceRawTextHelpFormatter(
    argparse.RawDescriptionHelpFormatter
):
    """A help text formatter allowing more customization of newlines."""

    def __add_whitespace(self, idx: int, iWSpace: int, text: str) -> str:
        if idx == 0:
            return text
        return (" " * iWSpace) + text

    def _split_lines(self, text: str, width: int) -> t.List[str]:
        textRows = text.splitlines()
        for idx, line in enumerate(textRows):
            search = re.search(r"\s*[0-9\-]{0,}\.?\s*", line)
            if line.strip() == "":
                textRows[idx] = " "
            elif search:
                lWSpace = search.end()
                lines = [
                    self.__add_whitespace(i, lWSpace, x)
                    for i, x in enumerate(textwrap.wrap(line, width))
                ]
                textRows[idx] = lines  # type: ignore

        return [item for sublist in textRows for item in sublist]
